import dataclasses

import numpy as np
import shapely

from threedigrid_builder.base import Lines, Nodes, OutputInterface
from threedigrid_builder.constants import (
    BoundaryType,
    CalculationType,
    ContentType,
    LineType,
    Material,
    NodeType,
)
from threedigrid_builder.grid import Grid, GridMeta, PotentialBreaches

__all__ = ["DictOut"]


NODE_FIELDS = (
    "id",
    "node_type",
    "calculation_type",
    "content_type",
    "content_pk",
    "dmax",  # bottom level
    "dimp",  # bottom level for groundwater
    "storage_area",
    "boundary_id",  # referring to the id of the boundary condition
    "boundary_type",
    "manhole_id",  # referring to the id of a manhole
    "initial_waterlevel",
)


CELL_FIELDS = (
    "id",
    "node_type",
    "boundary_id",  # referring to the id of the boundary condition
    "boundary_type",
    "has_dem_averaged",  # Boolean attribute to tell if dem is averaged for node.
)

LINE_FIELDS = (
    "id",
    "kcu",  # calculation type of the line
    "node_1",  # node id
    "node_2",  # node id
    "ds1d",  # arclength
    "ds1d_half",
    "content_type",
    "content_pk",
    "dpumax",  # bottom_level at the velocity point
    "invert_level_start_point",  # bottom level at line start
    "invert_level_end_point",  # bottom level at line end
    "flod",  # obstacle height
    "cross_id1",  # the id of the cross section definition
    "cross_id2",  # the id of the cross section definition
    "cross_weight",
    "frict_type1",
    "frict_type2",
    "frict_value1",
    "frict_value2",
    "discharge_coefficient_positive",
    "discharge_coefficient_negative",
)


EMBEDDED_NODE_FIELDS = (
    "id",
    "content_type",
    "content_pk",
    "embedded_in",
)


META_FIELDS = (
    "epsg_code",
    "threedigrid_builder_version",
    "has_1d",
    "has_2d",
    "has_embedded",
    "has_breaches",
    "has_groundwater",
    "has_groundwater_flow",
    "has_interception",
    "has_pumpstations",
    "has_simple_infiltration",
    "has_max_infiltration_capacity",
    "has_interflow",
    "has_initial_waterlevels",
)


def _enum_to_str(arr, enum_type):
    result = np.full_like(arr, "", dtype=object)
    result[arr != -9999] = [enum_type(x).name for x in arr[arr != -9999]]
    return result


def increase(arr):
    """Increase arr by one where arr is not -9999"""
    return np.add(arr, 1 * (arr != -9999), dtype=arr.dtype)


class DictOut(OutputInterface):
    def __init__(self, path):
        if path is not None:
            raise ValueError("The dict writer cannot write to a path")
        super().__init__(path)

    @staticmethod
    def available():
        return True

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def write(self, grid: Grid, geometry_format="wkb"):
        """Write a grid to a dictionary of layers.

        A layer is a dictionary of 1D ndarrays. The following layers are returned:

        - nodes: all computational nodes as Point geometries.
        - cells: a subselection of the nodes as Polygon geometries
        - lines: all flowlines as LineString geometries. lines interconnect nodes.
        - breaches: potential breach locations (Point geometries)
        - nodes_embedded: (virtual) nodes that are embedded into computational nodes
        - meta: metadata (e.g. "epsg_code"). the fields are scalars, not 1D ndarrays

        Args:
            grid (Grid)
            geometry_format {"native", "wkb", "wkt"}: optionally serialize geometries

        Returns:
            dict of dicts of 1D ndarrays
        """
        if geometry_format == "native":
            geom_serializer = None
        elif geometry_format == "wkb":
            geom_serializer = shapely.to_wkb
        elif geometry_format == "wkt":
            geom_serializer = shapely.to_wkt
        else:
            raise ValueError(f"Unknown geometry format '{geometry_format}'")

        nodes, cells = self.get_nodes_cells(grid.nodes)
        nodes_emb = self.get_embedded(grid.nodes_embedded)
        lines = self.get_lines(grid.lines)
        breaches = self.get_breaches(grid.breaches)
        meta = self.get_meta(grid.meta)
        result = {
            "nodes": nodes,
            "cells": cells,
            "lines": lines,
            "breaches": breaches,
            "nodes_embedded": nodes_emb,
            "meta": meta,
        }
        if geom_serializer is not None:
            for key in result.keys():
                if result[key] is None or "geometry" not in result[key]:
                    continue
                result[key]["geometry"] = geom_serializer(result[key]["geometry"])
        return result

    def get_nodes_cells(self, nodes: Nodes):
        """Create "nodes" and "cells" dictionaries

        Args:
            nodes (Nodes)

        Returns:
            tuple 2 dicts of 1D ndarrays (nodes, cells)
        """
        node_data = nodes.to_dict()

        is_2d = np.isin(
            nodes.node_type, (NodeType.NODE_2D_OPEN_WATER, NodeType.NODE_2D_BOUNDARIES)
        )

        # construct points from nodes.coordinates
        node_geometries = np.empty(len(nodes), dtype=object)
        coordinates = node_data.pop("coordinates")
        has_coord = np.isfinite(coordinates).all(axis=1)
        node_geometries[has_coord] = shapely.points(coordinates[has_coord])

        # construct cells from nodes.bounds
        bounds = node_data.pop("bounds")[is_2d]
        has_cell = np.isfinite(bounds).all(axis=1)
        cell_geometries = np.empty(bounds.shape[0], dtype=object)
        cell_geometries[has_cell] = shapely.box(*bounds[has_cell].T)

        # convert enums to strings
        node_data["node_type"] = _enum_to_str(node_data["node_type"], NodeType)
        node_data["content_type"] = _enum_to_str(node_data["content_type"], ContentType)
        node_data["calculation_type"] = _enum_to_str(
            node_data["calculation_type"], CalculationType
        )
        node_data["boundary_type"] = _enum_to_str(
            node_data["boundary_type"], BoundaryType
        )

        # go from 0-based to 1-based indexing
        for field in ("id", "embedded_in"):
            node_data[field] = increase(node_data[field])

        node_data_filt = {field: node_data[field] for field in NODE_FIELDS}
        node_data_filt["geometry"] = node_geometries

        cell_data = {field: node_data[field][is_2d] for field in CELL_FIELDS}
        cell_data["geometry"] = cell_geometries

        return node_data_filt, cell_data

    def get_embedded(self, nodes_embedded: Nodes):
        """Creates the "nodes_embedded" dictionary

        Args:
            nodes_embedded (Nodes)

        Returns:
            dicts of 1D ndarrays
        """
        if nodes_embedded is None or len(nodes_embedded) == 0:
            return None
        node_data = nodes_embedded.to_dict()

        # construct points from nodes.coordinates
        node_geometries = np.empty(len(nodes_embedded), dtype=object)
        coordinates = node_data.pop("coordinates")
        has_coord = np.isfinite(coordinates).all(axis=1)
        node_geometries[has_coord] = shapely.points(coordinates[has_coord])

        # convert enums to strings
        node_data["node_type"] = _enum_to_str(node_data["node_type"], NodeType)
        node_data["content_type"] = _enum_to_str(node_data["content_type"], ContentType)
        node_data["calculation_type"] = _enum_to_str(
            node_data["calculation_type"], CalculationType
        )
        node_data["boundary_type"] = _enum_to_str(
            node_data["boundary_type"], BoundaryType
        )

        # go from 0-based to 1-based indexing
        for field in ("id", "embedded_in"):
            node_data[field] = increase(node_data[field])

        node_data_filt = {field: node_data[field] for field in EMBEDDED_NODE_FIELDS}
        node_data_filt["geometry"] = node_geometries

        return node_data_filt

    def get_lines(self, lines: Lines):
        """Get "lines" dictionary

        Args:
            lines (Lines)

        Returns:
            lines dict of 1D ndarrays
        """
        line_data = lines.to_dict()

        # convert enums to strings
        line_data["kcu"] = _enum_to_str(line_data["kcu"], LineType)
        line_data["content_type"] = _enum_to_str(line_data["content_type"], ContentType)

        # go from 0-based to 1-based indexing
        for field in ("id", "line"):
            line_data[field] = increase(line_data[field])

        # cast lines.line to 2 1D arrays
        line_data["node_1"], line_data["node_2"] = line_data.pop("line").T

        geometries = line_data["line_geometries"]
        line_data = {field: line_data[field] for field in LINE_FIELDS}
        line_data["geometry"] = geometries
        return line_data

    def get_breaches(self, breaches: PotentialBreaches):
        """Get "breaches" dictionary

        Args:
            breaches (Breaches)

        Returns:
            breaches dict of 1D ndarrays
        """
        if len(breaches) == 0:
            return

        # sort by id
        return {
            "id": increase(breaches.id),
            "line_id": increase(breaches.line_id),
            "content_pk": breaches.content_pk,
            "maximum_breach_depth": breaches.maximum_breach_depth,
            "levee_material": _enum_to_str(breaches.levee_material, Material),
            "geometry": breaches.the_geom,
            "code": breaches.code,
            "display_name": breaches.display_name,
        }

    def get_meta(self, meta: GridMeta):
        meta_data = dataclasses.asdict(meta)
        meta_data = {field: meta_data[field] for field in META_FIELDS}
        return meta_data



import itertools
from unittest import mock

import numpy as np
import pytest
import shapely
from numpy.testing import assert_array_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import (
    BoundaryType,
    CalculationType,
    ContentType,
    LineType,
    NodeType,
)
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import BoundaryConditions1D, BoundaryConditions2D, Grid


@pytest.fixture
def grid1d():
    return Grid(
        nodes=Nodes(
            id=[1, 2, 3, 4],
            calculation_type=CalculationType.ISOLATED,
            content_type=ContentType.TYPE_V2_CONNECTION_NODES,
            node_type=NodeType.NODE_1D_NO_STORAGE,
            content_pk=[1, 5, 9, 10],
        ),
        lines=Lines(
            id=[1, 2, 3],
            line=[(2, 1), (1, 3), (1, 4)],
            kcu=LineType.LINE_1D_ISOLATED,
        ),
    )


def test_1d_apply(grid1d):
    boundary_conditions_1d = BoundaryConditions1D(
        id=[10, 22],
        boundary_type=[2, 3],
        connection_node_id=[9, 5],
    )
    boundary_conditions_1d.apply(grid1d)

    # note that apply does not reorder the nodes, this happens later at grid.sort()
    assert_array_equal(grid1d.nodes.boundary_id, [-9999, 22, 10, -9999])
    assert_array_equal(grid1d.nodes.boundary_type, [-9999, 3, 2, -9999])
    assert_array_equal(grid1d.nodes.node_type[[0, 3]], NodeType.NODE_1D_NO_STORAGE)
    assert_array_equal(grid1d.nodes.node_type[[1, 2]], NodeType.NODE_1D_BOUNDARIES)
    assert_array_equal(grid1d.nodes.calculation_type[[0, 3]], CalculationType.ISOLATED)
    assert_array_equal(
        grid1d.nodes.calculation_type[[1, 2]], CalculationType.BOUNDARY_NODE
    )
    assert_array_equal(grid1d.lines.is_1d_boundary[[0, 1]], 1)
    assert_array_equal(grid1d.lines.is_1d_boundary[[2]], -9999)


def test_1d_connection_node_does_not_exist(grid1d):
    boundary_conditions_1d = BoundaryConditions1D(id=[1], connection_node_id=[6])
    with pytest.raises(
        SchematisationError, match=r".*\[1\] refer to missing connection nodes \[6\].*"
    ):
        boundary_conditions_1d.apply(grid1d)


def test_1d_connection_node_has_no_lines(grid1d):
    boundary_conditions_1d = BoundaryConditions1D(id=[1], connection_node_id=[5])
    grid1d.lines = grid1d.lines[[1, 2]]
    with pytest.raises(
        SchematisationError,
        match=r".*\[1\] refer to connection nodes that have no attached objects \(\[5\]\).*",
    ):
        boundary_conditions_1d.apply(grid1d)


def test_1d_connection_node_has_too_many_lines(grid1d):
    boundary_conditions_1d = BoundaryConditions1D(id=[1], connection_node_id=[1])
    with pytest.raises(
        SchematisationError,
        match=r".*one attached object is allowed for connection nodes \[1\].*",
    ):
        boundary_conditions_1d.apply(grid1d)


def test_1d_boundary_conditions_too_close(grid1d):
    boundary_conditions_1d = BoundaryConditions1D(id=[1, 2], connection_node_id=[1, 5])
    grid1d.lines = grid1d.lines[[0]]
    with pytest.raises(SchematisationError, match=r".*\[1, 2\] are too close to.*"):
        boundary_conditions_1d.apply(grid1d)


@pytest.fixture
def grid2d():
    # A grid as follows, without lines.
    #
    #       + - + - - - +
    #       | 5 |       |
    #   + - + - +   2   |
    #   | 3 | 4 |       |
    #   + - + - + - + - +
    #   |       | 7 | 8 |
    #   |   1   + - + - +
    #   |       | 6 |
    #   + - - - + - +
    #
    pixel_size = 5.0
    min_cell_size = 20.0  # the size of the smallest cell
    origin = 46.0, 47.0  # coordinate of the bottomleft corner
    nodk = np.array([2, 2, 1, 1, 1, 1, 1, 1])
    nodm = np.array([1, 2, 1, 2, 2, 3, 3, 4])
    nodn = np.array([1, 2, 3, 3, 4, 1, 2, 2])
    _size = 2 ** (nodk - 1) * min_cell_size
    bounds = np.empty((len(nodk), 4))
    bounds[:, 0] = (nodm - 1) * _size + origin[0]
    bounds[:, 1] = (nodn - 1) * _size + origin[1]
    bounds[:, 2] = bounds[:, 0] + _size
    bounds[:, 3] = bounds[:, 1] + _size
    pixel_coords = np.empty((len(nodk), 4), dtype=int)
    pixel_coords[:, 0] = (nodm - 1) * _size / pixel_size
    pixel_coords[:, 1] = (nodn - 1) * _size / pixel_size
    pixel_coords[:, 2] = pixel_coords[:, 0] + _size / pixel_size
    pixel_coords[:, 3] = pixel_coords[:, 1] + _size / pixel_size
    nodes = Nodes(
        id=range(1, 9),
        node_type=NodeType.NODE_2D_OPEN_WATER,
        bounds=bounds,
        pixel_coords=pixel_coords,
        nodk=nodk,
        nodm=nodm,
        nodn=nodn,
    )
    grid = Grid(nodes=nodes, lines=Lines(id=[]))
    grid.quadtree = mock.Mock()
    grid.quadtree.quad_idx = np.array(
        [
            [0, 5, 2, 2],
            [3, 4, 2, 2],
            [1, 1, 7, 8],
            [1, 1, 6, 0],
        ]
    )[::-1].T
    return grid


# some expected boundary cells
bottom_1 = [46, 7, 86, 47]
bottom_6 = [86, 27, 106, 47]
left_1 = [6, 47, 46, 87]
left_3 = [26, 87, 46, 107]
top_2 = [86, 127, 126, 167]
top_5 = [66, 127, 86, 147]
right_2 = [126, 87, 166, 127]
right_8 = [126, 67, 146, 87]

SOUTH = LineType.LINE_2D_BOUNDARY_SOUTH
NORTH = LineType.LINE_2D_BOUNDARY_NORTH
EAST = LineType.LINE_2D_BOUNDARY_EAST
WEST = LineType.LINE_2D_BOUNDARY_WEST


@pytest.mark.parametrize(
    "bc_coords, bounds, boundary_id, nodm, nodn, kcu, line, cross_pix_coords",
    [
        (  # Boundary condition exactly at the bottom
            [[(46, 47), (106, 47)]],
            (bottom_1, bottom_6),
            0,
            [1, 3],
            [0, 0],
            SOUTH,
            [(9, 1), (10, 6)],
            [(0, 0, 8, 0), (8, 0, 12, 0)],
        ),
        (  # Boundary condition exactly at the left
            [[(46, 47), (46, 107)]],
            (left_1, left_3),
            0,
            [0, 0],
            [1, 3],
            WEST,
            [(9, 1), (10, 3)],
            [(0, 0, 0, 8), (0, 8, 0, 12)],
        ),
        (  # Boundary condition exactly at the top
            [[(66, 127), (126, 127)]],
            (top_2, top_5),
            0,
            [2, 2],
            [3, 5],
            NORTH,
            [(2, 9), (5, 10)],
            [(8, 16, 16, 16), (4, 16, 8, 16)],
        ),
        (  # Boundary condition exactly at the right
            [[(126, 67), (126, 127)]],
            (right_2, right_8),
            0,
            [3, 5],
            [2, 2],
            EAST,
            [(2, 9), (8, 10)],
            [(16, 8, 16, 16), (16, 4, 16, 8)],
        ),
        (  # Boundary condition exactly at the bottom and then one left (with overlap)
            [[(46, 47), (106, 47)], [(46, 47), (46, 107)]],
            (bottom_1, bottom_6, left_3),
            (0, 0, 1),
            [1, 3, 0],
            [0, 0, 3],
            [SOUTH, SOUTH, WEST],
            [(9, 1), (10, 6), (11, 3)],
            [(0, 0, 8, 0), (8, 0, 12, 0), (0, 8, 0, 12)],
        ),
        (  # Boundary condition exactly at the left and then one bottom (with overlap)
            [[(46, 47), (46, 107)], [(46, 47), (106, 47)]],
            (left_1, left_3, bottom_6),
            (0, 0, 1),
            [0, 0, 3],
            [1, 3, 0],
            [WEST, WEST, SOUTH],
            [(9, 1), (10, 3), (11, 6)],
            [(0, 0, 0, 8), (0, 8, 0, 12), (8, 0, 12, 0)],
        ),
        (  # Boundary condition just outside of the model, bottom
            [[(46, 40), (106, 41)]],
            (bottom_1, bottom_6),
            0,
            [1, 3],
            [0, 0],
            SOUTH,
            [(9, 1), (10, 6)],
            [(0, 0, 8, 0), (8, 0, 12, 0)],
        ),
        (  # Boundary condition just outside of the model, left
            [[(40, 47), (41, 107)]],
            (left_1, left_3),
            0,
            [0, 0],
            [1, 3],
            WEST,
            [(9, 1), (10, 3)],
            [(0, 0, 0, 8), (0, 8, 0, 12)],
        ),
        (  # Boundary condition just outside of the model, top
            [[(66, 130), (126, 132)]],
            (top_2, top_5),
            0,
            [2, 2],
            [3, 5],
            NORTH,
            [(2, 9), (5, 10)],
            [(8, 16, 16, 16), (4, 16, 8, 16)],
        ),
        (  # Boundary condition just outside of the model, right
            [[(127, 67), (126, 127)]],
            (right_2, right_8),
            0,
            [3, 5],
            [2, 2],
            EAST,
            [(2, 9), (8, 10)],
            [(16, 8, 16, 16), (16, 4, 16, 8)],
        ),
    ],
)
def test_2d_boundary_condition(
    grid2d,
    bc_coords,
    bounds,
    nodm,
    nodn,
    kcu,
    boundary_id,
    line,
    cross_pix_coords,
):
    boundary_conditions_2d = BoundaryConditions2D(
        id=range(len(bc_coords)),
        boundary_type=[3] * len(bc_coords),
        the_geom=shapely.linestrings(bc_coords),
    )

    nodes, lines = boundary_conditions_2d.get_nodes_and_lines(
        grid2d.nodes,
        grid2d.cell_tree,
        grid2d.quadtree,
        itertools.count(9),
        itertools.count(),
    )

    assert_array_equal(nodes.node_type, NodeType.NODE_2D_BOUNDARIES)
    assert_array_equal(nodes.boundary_id, boundary_id)
    assert_array_equal(nodes.boundary_type, 3)
    assert_array_equal(nodes.bounds, bounds)
    assert_array_equal(nodes.pixel_coords, -9999)
    assert_array_equal(nodes.nodm, nodm)
    assert_array_equal(nodes.nodn, nodn)
    assert_array_equal(lines.kcu, kcu)
    assert_array_equal(lines.line, line)
    assert_array_equal(lines.cross_pix_coords, cross_pix_coords)


@pytest.mark.parametrize(
    "bc_coords, expected_message",
    [
        ([[(46, 0), (106, 0)]], r".*does not touch any edge cell."),
        (
            [[(46, 77), (106, 77)]],
            r".*different edge coordinates \(y=\[47.0, 67.0\]\).",
        ),
        (
            [[(76, 47), (76, 107)]],
            r".*different edge coordinates \(x=\[46.0, 66.0\]\).",
        ),
    ],
)
def test_2d_boundary_condition_err(grid2d, bc_coords, expected_message):
    boundary_conditions_2d = BoundaryConditions2D(
        id=range(len(bc_coords)),
        boundary_type=[3] * len(bc_coords),
        the_geom=shapely.linestrings(bc_coords),
    )

    with pytest.raises(SchematisationError, match=expected_message):
        boundary_conditions_2d.get_nodes_and_lines(
            grid2d.nodes,
            grid2d.cell_tree,
            grid2d.quadtree,
            itertools.count(),
            itertools.count(),
        )


@pytest.fixture
def grid2d_gw():
    nodes = Nodes(
        id=range(4),
        node_type=([NodeType.NODE_2D_OPEN_WATER] * 2)
        + ([NodeType.NODE_2D_GROUNDWATER] * 2),
        bounds=[(2.0, 3.0, 7.0, 8.0), (7.0, 3.0, 12.0, 8.0)] * 2,
        pixel_coords=[(0, 0, 10, 10), (10, 0, 20, 10)] * 2,
        nodk=[1, 1, 1, 1],
        nodm=[1, 2, 1, 2],
        nodn=[1, 1, 1, 1],
    )
    grid = Grid(nodes=nodes, lines=Lines(id=[]))
    grid.quadtree = mock.Mock()
    grid.quadtree.quad_idx = np.array([[1, 2]])[::-1].T
    return grid


@pytest.mark.parametrize(
    "boundary_type,node_type,kcu,line",
    [
        (
            BoundaryType.WATERLEVEL,
            NodeType.NODE_2D_BOUNDARIES,
            LineType.LINE_2D_BOUNDARY_EAST,
            (1, 9),
        ),
        (
            BoundaryType.DISCHARGE,
            NodeType.NODE_2D_BOUNDARIES,
            LineType.LINE_2D_BOUNDARY_EAST,
            (1, 9),
        ),
        (
            BoundaryType.GROUNDWATERLEVEL,
            NodeType.NODE_2D_GROUNDWATER_BOUNDARIES,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST,
            (3, 9),
        ),
        (
            BoundaryType.GROUNDWATERDISCHARGE,
            NodeType.NODE_2D_GROUNDWATER_BOUNDARIES,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST,
            (3, 9),
        ),
    ],
)
def test_2d_boundary_condition_types(grid2d_gw, boundary_type, node_type, kcu, line):
    boundary_conditions_2d = BoundaryConditions2D(
        id=[1],
        boundary_type=[boundary_type],
        the_geom=[shapely.linestrings([(12.0, 3.0), (12.0, 8.0)])],
    )
    nodes, lines = boundary_conditions_2d.get_nodes_and_lines(
        grid2d_gw.nodes,
        grid2d_gw.cell_tree,
        grid2d_gw.quadtree,
        itertools.count(9),
        itertools.count(7),
    )

    assert_array_equal(nodes.id, [9])
    assert_array_equal(nodes.node_type, [node_type])
    assert_array_equal(nodes.boundary_id, [1])
    assert_array_equal(nodes.boundary_type, [boundary_type])
    assert_array_equal(nodes.bounds, [(12.0, 3.0, 17, 8.0)])
    assert_array_equal(nodes.pixel_coords, -9999)
    assert_array_equal(nodes.nodm, [3])
    assert_array_equal(nodes.nodn, [1])
    assert_array_equal(lines.id, [7])
    assert_array_equal(lines.kcu, [kcu])
    assert_array_equal(lines.line, [line])
    assert_array_equal(lines.cross_pix_coords, [(20, 0, 20, 10)])


def test_2d_boundary_condition_combined(grid2d_gw):
    boundary_conditions_2d = BoundaryConditions2D(
        id=[1, 2],
        boundary_type=[BoundaryType.WATERLEVEL, BoundaryType.GROUNDWATERLEVEL],
        the_geom=[shapely.linestrings([(12.0, 3.0), (12.0, 8.0)])] * 2,
    )

    nodes, lines = boundary_conditions_2d.get_nodes_and_lines(
        grid2d_gw.nodes,
        grid2d_gw.cell_tree,
        grid2d_gw.quadtree,
        itertools.count(9),
        itertools.count(7),
    )

    assert_array_equal(nodes.id, [9, 10])
    assert_array_equal(
        nodes.node_type,
        [NodeType.NODE_2D_BOUNDARIES, NodeType.NODE_2D_GROUNDWATER_BOUNDARIES],
    )
    assert_array_equal(nodes.boundary_id, [1, 2])
    assert_array_equal(
        nodes.boundary_type, [BoundaryType.WATERLEVEL, BoundaryType.GROUNDWATERLEVEL]
    )
    assert_array_equal(nodes.bounds, [(12.0, 3.0, 17, 8.0)] * 2)
    assert_array_equal(nodes.pixel_coords, -9999)
    assert_array_equal(nodes.nodm, 3)
    assert_array_equal(nodes.nodn, 1)
    assert_array_equal(lines.id, [7, 8])
    assert_array_equal(
        lines.kcu,
        [LineType.LINE_2D_BOUNDARY_EAST, LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST],
    )
    assert_array_equal(lines.line, [(1, 9), (3, 10)])
    assert_array_equal(lines.cross_pix_coords, [(20, 0, 20, 10)] * 2)



import itertools
from unittest import mock

import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_array_equal

from threedigrid_builder.constants import ContentType
from threedigrid_builder.grid import ConnectionNodes, Culverts, Orifices, Weirs


@pytest.fixture
def connection_nodes():
    # Used to map connection_node_start/end_id to an index (sequence id)
    return ConnectionNodes(
        id=np.array([21, 25, 33, 42]),
        the_geom=shapely.points([(0, 21), (1, 25), (2, 33), (3, 42)]),
    )


@pytest.fixture
def culverts():
    return Culverts(
        id=[1, 2],
        dist_calc_points=[5.0, np.nan],
        code=["one", "two"],
        connection_node_start_id=[21, 21],
        connection_node_end_id=[25, 42],
        calculation_type=[2, 1],
        invert_level_start_point=[3.0, 5.0],
        invert_level_end_point=[4.0, 6.0],
        the_geom=[
            shapely.linestrings([(0, 21), (0.5, 22), (1, 25)]),
            shapely.linestrings([(0, 21), (3, 42)]),
        ],
    )


@pytest.fixture(scope="module", params=[Weirs, Orifices])
def two_weir_orifices(request):
    return request.param(
        id=[1, 2],
        connection_node_start_id=[21, 25],
        connection_node_end_id=[42, 33],
        crest_level=[2.3, 4.5],
        crest_type=[4, 3],
        friction_type=[1, 1],
        friction_value=[31, 41],
        cross_section_definition_id=[17, 18],
    )


def test_get_lines(connection_nodes, two_weir_orifices):
    lines = two_weir_orifices.get_lines(
        connection_nodes,
        itertools.count(start=0),
        connection_node_offset=100,
    )

    if two_weir_orifices.__class__ is Weirs:
        expected_content_type = ContentType.TYPE_V2_WEIR
    elif two_weir_orifices.__class__ is Orifices:
        expected_content_type = ContentType.TYPE_V2_ORIFICE

    assert_array_equal(lines.id, range(2))
    assert_array_equal(lines.line, [(100, 103), (101, 102)])
    assert_array_equal(lines.content_type, expected_content_type)
    assert_array_equal(lines.content_pk, [1, 2])
    assert_array_equal(lines.kcu, [4, 3])
    assert_array_equal(lines.cross_id1, [17, 18])
    assert_array_equal(lines.cross_id2, [17, 18])
    assert_array_equal(lines.cross_weight, [1.0, 1.0])
    assert_array_equal(lines.dpumax, [2.3, 4.5])
    assert_array_equal(lines.invert_level_start_point, [2.3, 4.5])
    assert_array_equal(lines.invert_level_end_point, [2.3, 4.5])
    assert_array_equal(lines.frict_type1, [1, 1])
    assert_array_equal(lines.frict_type2, [1, 1])
    assert_array_equal(lines.frict_value1, [31, 41])
    assert_array_equal(lines.frict_value2, [31, 41])


@pytest.mark.parametrize(
    "culvert_ids,ds,expected",
    [
        ([1], [5.0], [3.5]),
        ([2], [0.5], [5.5]),
        ([1, 2], [5.0, 0.5], [3.5, 5.5]),
        ([2, 1], [0.5, 5.0], [5.5, 3.5]),
        ([1, 1, 2, 2], [5.0, 7.5, 0.5, 0.25], [3.5, 3.75, 5.5, 5.25]),
    ],
)
def test_culverts_compute_bottom_level(culvert_ids, ds, culverts, expected):
    # set geometries with lengths 10 and 1 (resp. id 1 and 2)
    # invert levels are [3, 4] for id=1 and [5, 6] for id=2
    culverts.the_geom = shapely.linestrings([[(0, 0), (0, 10)], [(2, 2), (3, 2)]])

    actual = culverts.compute_bottom_level(culvert_ids, ds)

    assert_almost_equal(actual, expected)


def test_culverts_compute_bottom_level_raises_no_geom(culverts):
    culverts.the_geom[:] = None
    with pytest.raises(ValueError, match=".*Call set_geometries first.*"):
        culverts.compute_bottom_level([1], [0.2])


@pytest.mark.parametrize(
    "culvert_ids,ds",
    [
        ([1], [10.1]),
        ([2], [-1e-7]),
    ],
)
def test_culverts_compute_bottom_level_raises_out_of_bounds(culvert_ids, ds, culverts):
    culverts.the_geom = shapely.linestrings([[(0, 0), (0, 10)], [(2, 2), (3, 2)]])
    with pytest.raises(ValueError, match=".*outside of the linear object bounds.*"):
        culverts.compute_bottom_level(culvert_ids, ds)


def test_culverts_get_1d2d_exchange_levels(culverts):
    content_pk = np.array([1, 1, 2])
    s1d = np.array([12.0, 13.0, 15.0])
    connection_nodes = mock.Mock()

    with mock.patch.object(culverts, "compute_drain_level") as compute_drain_level:
        compute_drain_level.return_value = np.array([1.8, 2.5, 3.2])
        actual = culverts.get_1d2d_exchange_levels(
            content_pk=content_pk, s1d=s1d, connection_nodes=connection_nodes
        )

        _, kwargs = compute_drain_level.call_args
        assert_array_equal(kwargs["ids"], content_pk)
        assert_array_equal(kwargs["s"], s1d)
        assert kwargs["connection_nodes"] is connection_nodes

    # interpolation between manhole drain levels is further tested elsewhere
    assert_array_equal(actual, compute_drain_level.return_value)


def test_is_closed(culverts):
    actual = culverts.is_closed(np.array([1, 3]))

    assert_array_equal(actual, [True, True])



from unittest import mock

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

from threedigrid_builder.constants import CrossSectionShape
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import CrossSectionDefinitions, CrossSections
from threedigrid_builder.grid.cross_section_definitions import (
    tabulate_builtin,
    tabulate_closed_rectangle,
    tabulate_egg,
    tabulate_inverted_egg,
    tabulate_tabulated,
    tabulate_yz,
)

SHP = CrossSectionShape


@pytest.fixture
def cross_section_definitions():
    return CrossSectionDefinitions(
        id=[1, 3, 9],
        code=["1", "3", "9"],
        shape=[SHP.CIRCLE, SHP.TABULATED_TRAPEZIUM, SHP.TABULATED_RECTANGLE],
        width=["1.22", "3.7 5.0", "5.0 6.0"],
        height=["", "1 2", "1 2"],
    )


def test_convert_multiple(cross_section_definitions):
    table_1 = np.random.random((9, 2))
    table_2 = np.random.random((4, 2))
    with mock.patch.dict(
        "threedigrid_builder.grid.cross_section_definitions.tabulators",
        {
            SHP.CIRCLE: mock.Mock(return_value=(1, 0.1, None, None)),
            SHP.TABULATED_TRAPEZIUM: mock.Mock(return_value=(5, 15.0, 2.0, table_1)),
            SHP.TABULATED_RECTANGLE: mock.Mock(return_value=(6, 11.0, 2.0, table_2)),
        },
    ):
        actual = cross_section_definitions.convert([1, 3, 9])

        assert len(actual) == 3
        assert_array_equal(actual.id, [0, 1, 2])
        assert_array_equal(actual.content_pk, cross_section_definitions.id)
        assert_array_equal(actual.code, cross_section_definitions.code)
        assert_array_equal(actual.shape, [1, 5, 6])  # see mocks
        assert_array_equal(actual.width_1d, [0.1, 15.0, 11.0])  # see mocks
        assert_array_equal(actual.height_1d, [np.nan, 2.0, 2.0])
        assert_array_equal(actual.offset, [0, 0, 9])
        assert_array_equal(actual.count, [0, 9, 4])
        assert_array_equal(actual.tables, np.concatenate([table_1, table_2], axis=0))


def test_convert_multiple_filtered(cross_section_definitions):
    with mock.patch.dict(
        "threedigrid_builder.grid.cross_section_definitions.tabulators",
        {
            SHP.CIRCLE: mock.Mock(return_value=(1, 0.1, 0.1, None)),
        },
    ):
        actual = cross_section_definitions.convert([1])

        assert len(actual) == 1
        assert_array_equal(actual.id, [0])
        assert_array_equal(actual.content_pk, [1])


def test_convert_nonexisting_id(cross_section_definitions):
    with pytest.raises(
        SchematisationError,
        match=r"One or more objects refer to non-existing cross section definitions \[2, 4\].",
    ):
        cross_section_definitions.convert([1, 2, 3, 4])


def test_convert_empty(cross_section_definitions):
    result = cross_section_definitions.convert([])
    assert isinstance(result, CrossSections)
    assert len(result) == 0


def test_tabulate_builtin():
    actual = tabulate_builtin("my-shape", "1.52", "1.33")
    assert actual == ("my-shape", 1.52, None, None)


def test_tabulate_closed_rectangle():
    shape, width_1d, height_1d, table = tabulate_closed_rectangle(
        "my-shape", "1.52", "5.2"
    )

    assert shape == CrossSectionShape.TABULATED_RECTANGLE
    assert width_1d == 1.52
    assert height_1d == 5.2
    assert_almost_equal(table, np.array([[0.0, 1.52], [5.2, 0.0]]))


def test_tabulate_egg():
    shape, width_1d, height_1d, table = tabulate_egg("my-shape", "1.52", "ignored")

    assert shape == CrossSectionShape.TABULATED_TRAPEZIUM
    assert width_1d == 1.52
    assert height_1d == 1.52 * 1.5

    # the expected table is exactly what inpy returns for a width of 1.52
    expected_table = np.array(
        [
            [0.0, 0.0],
            [0.152, 0.584],
            [0.304, 0.816],
            [0.456, 0.99],
            [0.608, 1.128],
            [0.76, 1.242],
            [0.912, 1.336],
            [1.064, 1.41],
            [1.216, 1.468],
            [1.368, 1.506],
            [1.52, 1.52],
            [1.672, 1.504],
            [1.824, 1.442],
            [1.976, 1.31],
            [2.128, 1.038],
            [2.28, 0.0],
        ]
    )
    assert_almost_equal(table, expected_table, decimal=3)


def test_tabulate_tabulated():
    shape, width_1d, height_1d, table = tabulate_tabulated("my-shape", "1 2 3", "0 1 2")

    assert shape == "my-shape"
    assert width_1d == 3.0  # the max
    assert height_1d == 2.0
    assert_almost_equal(table, np.array([[0, 1], [1, 2], [2, 3]], dtype=float))


@pytest.mark.parametrize(
    "width,height,match",
    [
        ("", "1", r"Unable to parse cross section definition.*"),
        ("1", "", r"Unable to parse cross section definition.*"),
        ("", "", r"Unable to parse cross section definition.*"),
        ("1", "1 2", r".*of tabulated or profile type must have equal number.*"),
        ("1 2", "1", r".*of tabulated or profile type must have equal number.*"),
        ("1 1", "2 1", r".*of tabulated type must have increasing heights.*"),
    ],
)
def test_tabulate_tabulated_err(width, height, match):
    with pytest.raises(SchematisationError, match=match):
        tabulate_tabulated(CrossSectionShape.TABULATED_RECTANGLE, width, height)


def test_tabulate_inverted_egg():
    shape, width_1d, height_1d, table = tabulate_inverted_egg(
        "my-shape", "1.52", "ignored"
    )

    assert shape == CrossSectionShape.TABULATED_TRAPEZIUM
    assert width_1d == 1.52
    assert height_1d == 1.52 * 1.5

    # the expected table is exactly what inpy returns for a width of 1.52
    expected_table = np.array(
        [
            [0.0, 0.0],
            [0.152, 1.038],
            [0.304, 1.31],
            [0.456, 1.442],
            [0.608, 1.504],
            [0.76, 1.52],
            [0.912, 1.506],
            [1.064, 1.468],
            [1.216, 1.41],
            [1.368, 1.336],
            [1.52, 1.242],
            [1.672, 1.128],
            [1.824, 0.99],
            [1.976, 0.816],
            [2.128, 0.584],
            [2.28, 0.0],
        ]
    )
    assert_almost_equal(table, expected_table, decimal=3)


@pytest.mark.parametrize(
    "width,height,exp_width,exp_height,exp_table",
    [
        ("0 0.5 1 1.5", "0.5 0 0 0.5", 1.5, 0.5, [[0, 0.5], [0.5, 1.5]]),
        (
            "0 0.5 1 1.5",
            "0.5 0 0 0.25",
            1.5,
            0.5,
            [[0, 0.5], [0.25, 1.25], [0.5, 1.5]],
        ),
        (
            "0 1 2 3 4 5",
            "1 0 0.5 0.5 0 1",
            5,
            1,
            [[0, 0], [0.5, 3], [0.5, 4], [1, 5]],
        ),
        (
            "0 1 2 2 0 0",
            "0.5 0 0.5 1.5 1.5 0.5",
            2.0,
            1.5,
            [[0, 0], [0.5, 2.0], [1.5, 2.0], [1.5, 0.0]],
        ),
        ("0 0.5 0.75 1.0 1.5", "0.5 0 0 0 0.5", 1.5, 0.5, [[0, 0.5], [0.5, 1.5]]),
        ("0 1 0 1 0", "0 1 1 0 0", 1, 1, [[0, 1], [0.5, 0], [1, 1], [1, 0]]),
    ],
)
def test_tabulate_yz(width, height, exp_width, exp_height, exp_table):
    shape, width_1d, height_1d, table = tabulate_yz("my-shape", width, height)

    assert shape == CrossSectionShape.TABULATED_TRAPEZIUM
    assert width_1d == exp_width
    assert height_1d == exp_height
    assert_almost_equal(table, np.array(exp_table, dtype=float))


@pytest.mark.parametrize(
    "width,height,match",
    [
        ("", "1", r"Unable to parse cross section definition.*"),
        ("1", "", r"Unable to parse cross section definition.*"),
        ("", "", r"Unable to parse cross section definition.*"),
        ("1", "1 2", r".*of tabulated or profile type must have equal number.*"),
        ("1 2", "1", r".*of tabulated or profile type must have equal number.*"),
        (
            "0 0.5 0",
            "0 1 0",
            r".*of closed profiles must have at least 4 coordinates.*",
        ),
        ("0 0.5", "0 1", r".*of open profiles must have at least 3 coordinates.*"),
        ("0 0.5 1.0", "0 1 -1", r".*cannot have negative height coordinate.*"),
        ("0 0.5 1.0", "1 2 1", r".*must have at least one height coordinate at 0.*"),
        ("0 0.5 1.0 0.5", "0 1 2 3", r".*should be closed or have increasing widths.*"),
    ],
)
def test_tabulate_yz_err(width, height, match):
    with pytest.raises(SchematisationError, match=match):
        tabulate_yz("my-shape", width, height)



from enum import Enum, IntEnum, unique

# The WKT spec to use in the output crs_wkt field:
WKT_VERSION = "WKT2_2015"


@unique
class CalculationType(IntEnum):
    # channels have +100 in the calculation type, this is mapped on read
    BOUNDARY_NODE = -1
    EMBEDDED = 0
    ISOLATED = 1
    CONNECTED = 2
    BROAD_CRESTED = 3  # only orifices + weirs, corresponds
    SHORT_CRESTED = 4  # only orifices + weirs
    DOUBLE_CONNECTED = 5


@unique
class NodeType(IntEnum):
    NODE_2D_OPEN_WATER = 1
    NODE_2D_GROUNDWATER = 2
    NODE_1D_NO_STORAGE = 3
    NODE_1D_STORAGE = 4
    NODE_2D_BOUNDARIES = 5
    NODE_2D_GROUNDWATER_BOUNDARIES = 6
    NODE_1D_BOUNDARIES = 7


@unique
class LineType(IntEnum):  # for kcu (calculation_type of a line)
    LINE_1D_EMBEDDED = int(CalculationType.EMBEDDED)
    LINE_1D_ISOLATED = int(CalculationType.ISOLATED)
    LINE_1D_CONNECTED = int(CalculationType.CONNECTED)
    LINE_1D_LONG_CRESTED = int(CalculationType.BROAD_CRESTED)
    LINE_1D_SHORT_CRESTED = int(CalculationType.SHORT_CRESTED)
    LINE_1D_DOUBLE_CONNECTED = int(CalculationType.DOUBLE_CONNECTED)
    LINE_1D2D_SINGLE_CONNECTED_CLOSED = 51
    LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER = 52
    LINE_1D2D_DOUBLE_CONNECTED_CLOSED = 53
    LINE_1D2D_DOUBLE_CONNECTED_OPEN_WATER = 54
    LINE_1D2D_POSSIBLE_BREACH = 55  # not used in gridbuilder
    LINE_1D2D_ACTIVE_BREACH = 56  # not used in gridbuilder
    LINE_1D2D_GROUNDWATER = 57
    LINE_2D_U = 98  # for internal use
    LINE_2D_V = 99  # for internal use
    LINE_2D = 100
    LINE_2D_OBSTACLE = 101
    LINE_2D_OBSTACLE_U = 102
    LINE_2D_OBSTACLE_V = 103
    LINE_2D_VERTICAL = 150
    LINE_2D_GROUNDWATER = -150
    LINE_2D_BOUNDARY_WEST = 200
    LINE_2D_BOUNDARY_EAST = 300
    LINE_2D_BOUNDARY_SOUTH = 400
    LINE_2D_BOUNDARY_NORTH = 500
    LINE_2D_GROUNDWATER_BOUNDARY_WEST = 600
    LINE_2D_GROUNDWATER_BOUNDARY_EAST = 700
    LINE_2D_GROUNDWATER_BOUNDARY_SOUTH = 800
    LINE_2D_GROUNDWATER_BOUNDARY_NORTH = 900


@unique
class ContentType(IntEnum):
    TYPE_V2_PIPE = 1
    TYPE_V2_CHANNEL = 2
    TYPE_V2_CULVERT = 3
    TYPE_V2_ORIFICE = 4
    TYPE_V2_WEIR = 5
    TYPE_V2_MANHOLE = 6
    TYPE_V2_PUMPSTATION = 7
    # TYPE_V2_LEVEE = 8
    TYPE_V2_1D_LATERAL = 9
    TYPE_V2_1D_BOUNDARY_CONDITIONS = 10
    TYPE_V2_2D_BOUNDARY_CONDITIONS = 11
    TYPE_V2_CONNECTION_NODES = 12
    TYPE_V2_BREACH = 13
    TYPE_V2_CROSS_SECTION_DEF = 14
    TYPE_V2_CROSS_SECTION_LOCATION = 15
    # TYPE_V2_ADDED_CALCULATION_POINT = 16
    TYPE_V2_WINDSHIELD = 17
    TYPE_V2_EXCHANGE_LINE = 18


@unique
class FrictionType(IntEnum):
    CHEZY = 1
    MANNING = 2  # pipes have 4 here in the input, but this is mapped to 2 on read


@unique
class CrossSectionShape(IntEnum):
    CLOSED_RECTANGLE = 0  # --> only user input, convert to tabulated rectangle
    RECTANGLE = 1
    CIRCLE = 2
    EGG = 3  # --> only user input, convert to tabulated trapezium
    TABULATED_RECTANGLE = 5
    TABULATED_TRAPEZIUM = 6
    TABULATED_YZ = 7  # --> only user input, convert to tabulated trapezium
    INVERTED_EGG = 8  # --> only user input, convert to tabulated trapezium


@unique
class InitializationType(IntEnum):
    MAX = 0  # file present
    MIN = 1  # file present
    AVERAGE = 2  # file present
    NO_AGG = 3  # file present
    GLOBAL = 9  # no file present, use global value


@unique
class BoundaryType(IntEnum):
    WATERLEVEL = 1
    VELOCITY = 2
    DISCHARGE = 3
    RIEMANN = 4
    SOMMERFELD = 5
    GROUNDWATERLEVEL = 6
    GROUNDWATERDISCHARGE = 7


@unique
class InfiltrationSurfaceOption(IntEnum):
    RAIN = 0
    WHOLE_SURFACE = 1
    WET_SURFACE = 2


@unique
class InterflowType(IntEnum):
    NO_INTERLFOW = 0
    LOCAL_DEEPEST_POINT_SCALED_POROSITY = 1
    GLOBAL_DEEPEST_POINT_SCALED_POROSITY = 2
    LOCAL_DEEPEST_POINT_CONSTANT_POROSITY = 3
    GLOBAL_DEEPEST_POINT_CONSTANT_POROSITY = 4


@unique
class SurfaceClass(str, Enum):
    GESLOTEN_VERHARDING = "gesloten verharding"
    OPEN_VERHARDING = "open verharding"
    HALF_VERHARD = "half verhard"
    ONVERHARD = "onverhard"
    PAND = "pand"


class Material(IntEnum):
    SAND = 1
    CLAY = 2


class InflowType(IntEnum):
    NO_INFLOW = 0
    IMPERVIOUS_SURFACE = 1
    SURFACE = 2



try:
    import fiona  # NOQA, see Github issue 265
    import geopandas
except ImportError:
    geopandas = None
from threedigrid_builder.base import OutputInterface
from threedigrid_builder.grid import Grid

from .dict_out import DictOut

__all__ = ["GeopackageOut"]


class GeopackageOut(OutputInterface):
    def __init__(self, path):
        if not self.available():
            raise ImportError("Cannot write to GPKG if geopandas is not available.")
        if not path.suffix.lower() == ".gpkg":
            raise ValueError("Extension should be .gpkg")
        super().__init__(path)

    @staticmethod
    def available():
        return geopandas is not None

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def write(self, grid: Grid):
        # use the DictOut writer to transform to dictionaries
        with DictOut(path=None) as out:
            layers = out.write(grid, geometry_format="native")

        # write the obtained layers using geopandas
        for name, layer in layers.items():
            if layer is None or len(layer.get("geometry", [])) == 0:
                continue
            df = geopandas.GeoDataFrame(layer, crs=grid.meta.crs_wkt)
            df.to_file(self.path, layer=name, driver="GPKG")



import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.grid import DemAverageAreas
from threedigrid_builder.grid.grid import Grid


@pytest.fixture
def grid():
    id = np.array([0, 1, 2, 3])
    calculation_type = np.array([-9999, -9999, -9999, -9999])
    bounds = np.array(
        [
            [0.0, 0.0, 1.0, 1.0],
            [0.0, 1.0, 1.0, 2.0],
            [1.0, 0.0, 2.0, 1.0],
            [1.0, 1.0, 2.0, 2.0],
        ],
        dtype=np.float64,
    )
    nodes = Nodes(id=id, calculation_type=calculation_type, bounds=bounds)
    grid = Grid(nodes=nodes, lines=Lines(id=np.array([1])))
    grid._cell_tree = shapely.STRtree(shapely.box(*grid.nodes.bounds.T))
    return grid


@pytest.fixture
def dem_average_areas():
    return DemAverageAreas(
        id=np.array([1], dtype=np.int32),
        the_geom=np.array([shapely.box(0.2, 0.2, 0.8, 1.8)]),
    )


@pytest.fixture
def dem_average_areas_no_intersect():
    return DemAverageAreas(
        id=np.array([1], dtype=np.int32),
        the_geom=np.array([shapely.box(11, 11, 18, 18)]),
    )


def test_dem_average_areas(grid, dem_average_areas):
    grid.set_dem_averaged_cells(dem_average_areas)
    assert_almost_equal(grid.nodes.has_dem_averaged, [1, 1, -9999, -9999])


def test_dem_average_areas_no_intersections(grid, dem_average_areas_no_intersect):
    grid.set_dem_averaged_cells(dem_average_areas_no_intersect)
    assert_almost_equal(grid.nodes.has_dem_averaged, [-9999, -9999, -9999, -9999])



import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_array_equal
from shapely.testing import assert_geometries_equal

from threedigrid_builder.base import Array, LinesOnLine, LineStrings, PointsOnLine


class LinearObject:
    id: int
    the_geom: shapely.Geometry


class LinearObjects(Array[LinearObject]):
    pass


@pytest.fixture
def linestrings():
    return LineStrings(
        LinearObjects(id=[0], the_geom=shapely.linestrings([[(0, 0), (6, 0), (6, 6)]]))
    )


@pytest.fixture
def two_linestrings():
    return LineStrings(
        LinearObjects(
            id=[1, 3],
            the_geom=np.array(
                [
                    shapely.linestrings([(0, 10), (10, 10)]),
                    shapely.linestrings([(0, 0), (6, 0), (6, 6)]),
                ]
            ),
        )
    )


@pytest.mark.parametrize(
    "s1d,expected_idx,expected_idx_s1d_start, expected_s1d_end",
    [
        ([], [0], [0.0], [12.0]),
        ([6.0], [0, 0], [0.0, 6.0], [6.0, 12.0]),
        ([2.0, 8.0], [0, 0, 0], [0.0, 2.0, 8.0], [2.0, 8.0, 12.0]),
    ],
)
def test_segmentize_one(
    linestrings, s1d, expected_idx, expected_idx_s1d_start, expected_s1d_end
):
    points_on_line = PointsOnLine(
        linestrings=linestrings,
        id=range(len(s1d)),
        s1d=s1d,
        linestring_idx=[0] * len(s1d),
    )
    lines = linestrings.segmentize(points_on_line)

    assert isinstance(lines, LinesOnLine)
    assert_array_equal(lines.linestring_idx, expected_idx)
    assert_array_equal(lines.s1d_start, expected_idx_s1d_start)
    assert_array_equal(lines.s1d_end, expected_s1d_end)


def test_segmentize_multiple(two_linestrings):
    points_on_line = PointsOnLine(
        linestrings=two_linestrings,
        id=[1, 2, 3],
        s1d=[5.0, 2.0, 8.0],
        linestring_idx=[0, 1, 1],
    )
    lines = two_linestrings.segmentize(points_on_line)

    assert isinstance(lines, LinesOnLine)
    assert_array_equal(lines.linestring_idx, [0, 0, 1, 1, 1])
    assert_array_equal(lines.s1d_start, [0.0, 5.0, 0.0, 2.0, 8.0])
    assert_array_equal(lines.s1d_end, [5.0, 10.0, 2.0, 8.0, 12.0])


@pytest.mark.parametrize(
    "geom,expected",
    [
        (None, None),
        ([(0, 21), (3, 42)], [(0, 21), (3, 42)]),
        ([(0, 21), (1, 2), (3, 42)], [(0, 21), (1, 2), (3, 42)]),
        ([(0, 21), (3, 42)][::-1], [(0, 21), (3, 42)]),
        ([(0, 21), (1, 2), (3, 42)][::-1], [(0, 21), (1, 2), (3, 42)]),
        ([(1, 21), (3, 41)], [(0, 21), (1, 21), (3, 41), (3, 42)]),
        ([(3, 41), (1, 21)], [(0, 21), (1, 21), (3, 41), (3, 42)]),
    ],
)
def test_sanitize(geom, expected):
    linestrings = LineStrings(
        LinearObjects(id=[0], the_geom=[shapely.linestrings(geom) if geom else None])
    )
    linestrings.sanitize(
        points_1=shapely.points([(0, 21)]), points_2=shapely.points([(3, 42)])
    )

    expected = [None] if expected is None else [shapely.linestrings(expected)]
    assert_geometries_equal(linestrings.the_geom, expected)


def test_point_on_line_the_geom_multiple(two_linestrings):
    points_on_line = PointsOnLine(
        linestrings=two_linestrings, id=[1, 2], s1d=[3.0, 4.0], linestring_idx=[0, 1]
    )
    actual = points_on_line.the_geom

    coords, index = shapely.get_coordinates(actual, return_index=True)

    assert_almost_equal(coords, [(3.0, 10.0), (4.0, 0)])
    assert_almost_equal(index, [0, 1])


def test_point_on_line_merge_with(linestrings):
    a = PointsOnLine(
        linestrings=linestrings,
        id=[0, 1],
        content_pk=[1, 2],
        s1d=[3.0, 4.0],
        linestring_idx=[0, 1],
    )
    b = PointsOnLine(
        linestrings=linestrings,
        id=[0, 1],
        content_pk=[5, 6],
        s1d=[1.0, 2.0],
        linestring_idx=[0, 1],
    )
    actual = a.merge_with(b)

    assert_array_equal(actual.s1d, [1.0, 3.0, 2.0, 4.0])
    assert_array_equal(actual.content_pk, [5, 1, 6, 2])
    assert_array_equal(actual.linestring_idx, [0, 0, 1, 1])


def test_point_on_line_at_start_at_end(two_linestrings):
    points_on_line = PointsOnLine(
        linestrings=two_linestrings,
        id=range(4),
        s1d=[0.0, 10.0, 0.0, 12.0],
        linestring_idx=[0, 0, 1, 1],
    )

    assert_array_equal(points_on_line.at_start, [True, False, True, False])
    assert_array_equal(points_on_line.at_end, [False, True, False, True])


def test_point_on_line_linestring_id(two_linestrings):
    points_on_line = PointsOnLine(
        linestrings=two_linestrings,
        id=range(3),
        linestring_idx=[0, 1, 0],
    )

    assert_array_equal(points_on_line.linestring_id, [1, 3, 1])


@pytest.mark.parametrize(
    "start,end,expected_coords",
    [
        (0.0, 12.0, [(0.0, 0.0), (6.0, 0.0), (6.0, 6.0)]),
        (1.45, 12.0, [(1.45, 0.0), (6.0, 0.0), (6.0, 6.0)]),
        (6.0, 12.0, [(6.0, 0.0), (6.0, 6.0)]),
        (6.7, 12.0, [(6.0, 0.7), (6.0, 6.0)]),
        (0.0, 11.5, [(0.0, 0.0), (6.0, 0.0), (6.0, 5.5)]),
        (0.0, 6.0, [(0.0, 0.0), (6.0, 0.0)]),
        (0.0, 3.5, [(0.0, 0.0), (3.5, 0.0)]),
        (1.0, 9.0, [(1.0, 0.0), (6.0, 0.0), (6.0, 3.0)]),
        (6.0 + 1e-7, 12.0, [(6.0 + 1e-7, 0.0), (6.0, 6.0)]),
        (6.0 + 1e-8, 12.0, [(6.0, 0.0), (6.0, 6.0)]),
        (6.0, 12.0 + 1e-15, [(6.0, 0.0), (6.0, 6.0)]),
    ],
)
def test_line_on_line_the_geom(linestrings, start, end, expected_coords):
    lines_on_line = LinesOnLine(
        linestrings=linestrings,
        id=[1],
        s1d_start=[start],
        s1d_end=[end],
        linestring_idx=[0],
    )
    geometries = lines_on_line.the_geom
    assert geometries.shape == (1,)
    assert_almost_equal(expected_coords, shapely.get_coordinates(geometries), decimal=7)


def test_line_on_line_the_geom_multiple(two_linestrings):
    lines_on_line = LinesOnLine(
        linestrings=two_linestrings,
        id=[1, 2],
        s1d_start=[3.0, 4.0],
        s1d_end=[9.0, 10.0],
        linestring_idx=[0, 1],
    )
    segments = lines_on_line.the_geom

    coords, index = shapely.get_coordinates(segments, return_index=True)

    assert_almost_equal(
        coords, [(3.0, 10.0), (9.0, 10.0), (4.0, 0), (6.0, 0.0), (6.0, 4.0)]
    )
    assert_almost_equal(index, [0, 0, 1, 1, 1])


@pytest.mark.parametrize(
    "segment_size,expected_size,expected_count",
    [
        (3.0, 12 / 4, 3),
        (4.0, 12 / 3, 2),
        (5.0, 12 / 2, 1),
        (6.0, 12 / 2, 1),
        (8.0, 12 / 2, 1),
        (9.0, 12, 0),
        (100.0, 12, 0),
    ],
)
def test_line_on_line_interpolate_points_one(
    linestrings, segment_size, expected_size, expected_count
):
    lines_on_line = LinesOnLine(
        linestrings=linestrings,
        id=[1],
        s1d_start=[0.0],
        s1d_end=linestrings.length,
        linestring_idx=[0],
    )
    points = lines_on_line.interpolate_points(segment_size)

    assert isinstance(points, PointsOnLine)
    assert len(points) == expected_count
    assert_array_equal(points.linestring_idx, 0)
    assert_almost_equal(points.s1d, np.arange(1, expected_count + 1) * expected_size)


@pytest.mark.parametrize(
    "s1d_start,s1d_end,expected_s1d",
    [
        (0.0, 6.0, [3.0]),
        (6.0, 12.0, [9.0]),
        (1.0, 10.0, [4.0, 7.0]),
    ],
)
def test_line_on_line_interpolate_points_partial(
    linestrings, s1d_start, s1d_end, expected_s1d
):
    lines_on_line = LinesOnLine(
        linestrings=linestrings,
        id=[1],
        s1d_start=[s1d_start],
        s1d_end=[s1d_end],
        linestring_idx=[0],
    )

    points = lines_on_line.interpolate_points(3.0)

    assert isinstance(points, PointsOnLine)
    assert_array_equal(points.linestring_idx, 0)
    assert_almost_equal(points.s1d, expected_s1d)


def test_line_on_line_interpolate_points_two(two_linestrings):
    lines_on_line = LinesOnLine(
        linestrings=two_linestrings,
        id=[1, 2],
        s1d_start=[0.0, 0.0],
        s1d_end=two_linestrings.length,
        linestring_idx=[0, 1],
    )

    points = lines_on_line.interpolate_points(5.0)

    assert isinstance(points, PointsOnLine)
    assert_array_equal(points.linestring_idx, [0, 1])
    assert_almost_equal(points.s1d, [5.0, 6.0])


def test_line_on_line_interpolate_points_two_partial(linestrings):
    lines_on_line = LinesOnLine(
        linestrings=linestrings,
        id=[1, 2],
        s1d_start=[0.0, 5.0],
        s1d_end=[5.0, 12.0],
        linestring_idx=[0, 0],
    )

    points = lines_on_line.interpolate_points(3.0)

    assert isinstance(points, PointsOnLine)
    assert_array_equal(points.linestring_idx, [0, 0])
    assert_almost_equal(points.s1d, [2.5, 8.5])


@pytest.mark.parametrize(
    "s1d_1, s1d_2, expected_1, expected_2",
    [
        ([6.0], [6.0], [0], [0]),
        ([6.0], [3.0], [0], [0]),
        ([6.0], [9.0], [0], [0]),
        ([3.0, 9.0], [0.0], [0], [1]),
        ([3.0, 9.0], [3.0], [0], [1]),
        ([3.0, 9.0], [6.0], [0], [1]),
        ([3.0, 9.0], [9.0], [0], [1]),
        ([3.0, 9.0], [12.0], [0], [1]),
        ([3.0, 6.0, 9.0], [0.0], [0], [1]),
        ([3.0, 6.0, 9.0], [3.0], [0], [1]),
        ([3.0, 6.0, 9.0], [4.0], [0], [1]),
        ([3.0, 6.0, 9.0], [6.0], [0], [1]),
        ([3.0, 6.0, 9.0], [7.0], [1], [2]),
        ([3.0, 6.0, 9.0], [9.0], [1], [2]),
        ([3.0, 6.0, 9.0], [12.0], [1], [2]),
        ([0.0, 6.0, 12.0], [0.0], [0], [1]),
        ([0.0, 6.0, 12.0], [12.0], [1], [2]),
    ],
)
def test_neighbours_one(linestrings, s1d_1, s1d_2, expected_1, expected_2):
    points_1 = PointsOnLine.from_s1d(linestrings, s1d_1, [0] * len(s1d_1))
    points_2 = PointsOnLine.from_s1d(linestrings, s1d_2, [0] * len(s1d_2))

    actual_1, actual_2 = points_1.neighbours(points_2)

    assert_array_equal(actual_1, expected_1)
    assert_array_equal(actual_2, expected_2)


@pytest.mark.parametrize(
    "s1d_1, s1d_2, expected_1, expected_2",
    [
        ([0.0, 10.0, 0.0, 12.0], [5.0, 6.0], [0, 2], [1, 3]),
        ([0.0, 10.0, 0.0, 12.0], [10.0, 0.0], [0, 2], [1, 3]),
        ([0.0, 10.0, 6.0, 12.0], [10.0, 0.0], [0, 2], [1, 3]),
        ([0.0, 5.0, 0.0, 12.0], [10.0, 0.0], [0, 2], [1, 3]),
    ],
)
def test_neighbours_multiple(two_linestrings, s1d_1, s1d_2, expected_1, expected_2):
    points_1 = PointsOnLine.from_s1d(two_linestrings, s1d_1, [0, 0, 1, 1])
    points_2 = PointsOnLine.from_s1d(two_linestrings, s1d_2, [0, 1])

    actual_1, actual_2 = points_1.neighbours(points_2)

    assert_array_equal(actual_1, expected_1)
    assert_array_equal(actual_2, expected_2)



import numpy as np
import shapely

from threedigrid_builder.base import Array
from threedigrid_builder.constants import CrossSectionShape
from threedigrid_builder.exceptions import SchematisationError

__all__ = ["CrossSectionDefinitions", "CrossSections"]


TABULATED_YZ_DECIMALS = 3


class CrossSectionDefinition:
    id: int
    code: str
    shape: CrossSectionShape
    height: str  # space-separated list of floats
    width: str  # space-separated list of floats


class CrossSectionDefinitions(Array[CrossSectionDefinition]):
    def convert(self, ids):
        """Convert to CrossSections.

        Args:
            ids (ndarray of int): A list of cross section definition ids to convert.
        """
        try:
            idx = self.id_to_index(ids, check_exists=True)
        except KeyError as e:
            if len(e.values) > 10:
                id_msg = str(e.values[:10].tolist()).replace("]", ", ...]")
            else:
                id_msg = str(e.values.tolist())
            raise SchematisationError(
                f"One or more objects refer to non-existing cross section definitions "
                f"{id_msg}."
            )

        result = CrossSections(
            id=range(len(idx)),
            content_pk=ids,
            code=self.code[idx],
            count=0,
        )
        if len(result) == 0:
            return result

        offset = 0
        tables = []

        # Numpy array views on width/height based on idx
        width_idx = self.width[idx]
        height_idx = self.height[idx]

        for i, shape in enumerate(self.shape[idx]):
            tabulator = tabulators[shape]
            result.shape[i], result.width_1d[i], result.height_1d[i], table = tabulator(
                shape, width_idx[i], height_idx[i]
            )
            if table is not None:
                result.count[i] = len(table)
                result.offset[i] = offset
                offset += len(table)
                tables.append(table)

        result.offset[:] = np.roll(np.cumsum(result.count), 1)
        result.offset[0] = 0

        if len(tables) > 0:
            result.tables = np.concatenate(tables, axis=0)
        else:
            result.tables = np.empty((0, 2))

        return result


class CrossSection:
    id: int
    code: str
    shape: CrossSectionShape
    content_pk: int
    width_1d: float
    height_1d: float
    offset: int
    count: int
    # tables: Tuple[float, float] has different length so is specified on CrossSections


class CrossSections(Array[CrossSection]):
    tables = None


def tabulate_builtin(shape, width, height):
    """Tabulate built-in shapes (rectangle, circle)

    Built-in geometries only require a width to be fully specified.

    Args:
        shape (CrossSectionShape): returned as is
        width (str): fully specifies the shape
        height (str): ignored

    Returns:
        tuple:  shape, width_1d (float), None, None
    """
    try:
        width = float(width)
    except ValueError:
        raise SchematisationError(
            f"Unable to parse cross section definition width (got: '{width}')."
        )

    return shape, width, None, None


def tabulate_egg(shape, width, height):
    """Tabulate the egg shape.

    Args:
        shape (CrossSectionShape): ignored
        width (str): the width of the egg, defines height and increment
        height (str): ignored; height is set to 1.5 * width

    Returns:
        tuple:  TABULATED_TRAPEZIUM, width_1d (float),
                height_1d (float), table (ndarray of shape (M, 2))
    """
    NUM_INCREMENTS = 16

    # width is the only constant; height derives from width
    try:
        width = float(width)
    except ValueError:
        raise SchematisationError(
            f"Unable to parse cross section definition width (got: '{width}')."
        )

    height = width * 1.5
    position = height / 3.0  # some parameter for the 'egg' curve
    heights = np.linspace(0, height, num=NUM_INCREMENTS, endpoint=True)

    # here comes the formulation of the "egg" curve, simplified from inpy
    h = height / 2
    w = width / 2
    a = position / 2
    x = h - heights
    p = (h**2 - (x**2)) * w**2
    q = h**2 + 2 * a * x + a**2
    widths = np.sqrt(p / q) * 2

    table = np.array([heights, widths]).T
    return CrossSectionShape.TABULATED_TRAPEZIUM, width, height, table


def tabulate_inverted_egg(shape, width, height):
    """Tabulate the egg shape, upside down.

    See tabulate_egg.
    """
    type_, width, height, table = tabulate_egg(shape, width, height)
    table[:, 1] = table[::-1, 1]
    return type_, width, height, table


def tabulate_closed_rectangle(shape, width, height):
    """Tabulate the closed rectangle shape.

    Args:
        shape (CrossSectionShape): ignored
        width (str): the width of the rectangle
        height (str): the height of the rectangle

    Returns:
        tuple:  TABULATED_RECTANGLE, width_1d (float),
                height (float), table (ndarray of shape (M, 2))
    """
    try:
        width = float(width)
        height = float(height)
    except ValueError:
        raise SchematisationError(
            f"Unable to parse cross section definition width and/or height "
            f"(got: '{width}', '{height}')."
        )
    table = np.array([[0.0, width], [height, 0.0]], order="F")
    return CrossSectionShape.TABULATED_RECTANGLE, width, height, table


def _parse_tabulated(width, height):
    try:
        heights = np.array([float(x) for x in height.split(" ")])
        widths = np.array([float(x) for x in width.split(" ")])
    except ValueError:
        raise SchematisationError(
            f"Unable to parse cross section definition width and/or height "
            f"(got: '{width}', '{height}')."
        )
    if len(heights) == 0:
        raise SchematisationError(
            f"Cross section definitions of tabulated or profile type must have at least one "
            f"height element (got: {height})."
        )
    if len(heights) != len(widths):
        raise SchematisationError(
            f"Cross section definitions of tabulated or profile type must have equal number of "
            f"height and width elements (got: {height}, {width})."
        )
    return widths, heights


def tabulate_tabulated(shape, width, height):
    """Tabulate the tabulated shapes.

    Args:
        shape (CrossSectionShape): returned as is
        width (str): space-separated widths
        height (str): space-separated heights

    Returns:
        tuple:  shape, width_1d (float),
                height_1d (float), table (ndarray of shape (M, 2))
    """
    widths, heights = _parse_tabulated(width, height)
    if len(heights) > 1 and np.any(np.diff(heights) < 0.0):
        raise SchematisationError(
            f"Cross section definitions of tabulated type must have increasing heights "
            f"(got: {height})."
        )

    return shape, np.max(widths), np.max(heights), np.array([heights, widths]).T


def tabulate_yz(shape, width, height):
    """Tabulate an (open or closed) YZ profile

    Args:
        shape: ignored
        width (str): space-separated horizontal (Y or X) coordinates
        height (str): space-separated Z coordinates

    Returns:
        tuple:  shape, width_1d (float),
                height_1d (float), table (ndarray of shape (M, 2))
    """
    ys, zs = _parse_tabulated(width, height)
    is_closed = ys[0] == ys[-1] and zs[0] == zs[-1]
    if is_closed and len(zs) < 4:
        raise SchematisationError(
            f"Cross section definitions of closed profiles must have at least "
            f"4 coordinates (got: {len(zs)})."
        )
    if (not is_closed) and len(zs) < 3:
        raise SchematisationError(
            f"Cross section definitions of open profiles must have at least "
            f"3 coordinates (got: {len(zs)})."
        )
    if np.any(zs < 0):
        raise SchematisationError(
            f"Cross section definitions of profiles cannot have negative "
            f"height coordinate (got: {height})."
        )
    if not np.any(zs == 0):
        raise SchematisationError(
            f"Cross section definitions of profiles must have at least one "
            f"height coordinate at 0.0 (got: {height})."
        )
    if not is_closed and np.any(np.diff(ys) < 0.0):
        raise SchematisationError(
            f"Cross section definition should be closed or have increasing widths "
            f"(got: {width})."
        )

    if is_closed:
        ys = ys[:-1]
        zs = zs[:-1]

    # Adapt non-unique height coordinates. Why?
    # Because if a segment of the profile is exactly horizontal, we need 2 widths
    seen = set()
    eps = 1 / (10 ** (TABULATED_YZ_DECIMALS + 1))
    for i, x in enumerate(zs):
        while x in seen:
            x += eps
        seen.add(x)
        zs[i] = x

    # shapely will automatically close an open profile
    profile = shapely.make_valid(shapely.polygons(np.array([ys, zs]).T))

    # take the length of the intersection with a horizontal line at each Z
    heights = np.unique(shapely.get_coordinates(profile)[:, 1])
    table = np.empty((len(heights), 2), dtype=float)
    y_min, y_max = ys.min(), ys.max()
    for i, height in enumerate(heights):
        line = shapely.linestrings([[y_min, height], [y_max, height]])
        cross_section_line = shapely.intersection(profile, line)
        width = shapely.length(cross_section_line)
        table[i, 0] = height
        table[i, 1] = width

    # For open profiles, if the end coordinate is closed, that means that left
    # and right sides are not equal in Z. We take the max width in that case.
    if not is_closed and table[-1, 1] == 0.0:
        table[-1, 1] = y_max - y_min

    # Eliminate duplicates and get rid of the epsilon introduced earlier
    # NB: Calccore allows a dicontinuity like [[0, 1], [1, 2], [1, 3]]
    table = np.round(table, decimals=TABULATED_YZ_DECIMALS)
    table = table[np.sort(np.unique(table, axis=0, return_index=True)[1])]

    # Drop first elements until we have 1 0.0 height at the start.
    # NB: An strong increase of width at the dry-wet transition may give issues.
    while True:
        if len(table) <= 1 or table[1, 0] > 0:
            break
        table = table[1:]

    height_1d, width_1d = table.max(axis=0).tolist()
    return CrossSectionShape.TABULATED_TRAPEZIUM, width_1d, height_1d, table


tabulators = {
    CrossSectionShape.CLOSED_RECTANGLE: tabulate_closed_rectangle,
    CrossSectionShape.RECTANGLE: tabulate_builtin,
    CrossSectionShape.CIRCLE: tabulate_builtin,
    CrossSectionShape.EGG: tabulate_egg,
    CrossSectionShape.TABULATED_RECTANGLE: tabulate_tabulated,
    CrossSectionShape.TABULATED_TRAPEZIUM: tabulate_tabulated,
    CrossSectionShape.TABULATED_YZ: tabulate_yz,
    CrossSectionShape.INVERTED_EGG: tabulate_inverted_egg,
}



import itertools
import logging
import math

import numpy as np

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import LineType, NodeType
from threedigrid_builder.exceptions import SchematisationError

from .fgrid import m_cells, m_quadtree

logger = logging.getLogger(__name__)


__all__ = ["QuadTree"]


def reduce_refinement_levels(refinements, num_refine_levels):
    """Determine the lowest refinement level that is actually in use.

    Args:
      refinements (GridRefinements): all gridrefinements polygon and linestrings
        with refinement levels. refinement_level may be changed inplace.
      num_refine_levels (int): the number of refinement levels supplied in the settings

    Returns:
      num_refine_levels (int), possibly lower than the input one
    """
    if refinements is not None and len(refinements) > 0:
        min_level = refinements.refinement_level.min().item()
    else:
        min_level = num_refine_levels

    min_level -= 1
    if refinements is not None:
        refinements.refinement_level -= min_level
    return num_refine_levels - min_level


class QuadTree:
    """Defines active cell levels for computational grid."""

    lgrmin: int
    pixel_size: float
    origin: float
    kmax: int
    mmax: int
    nmax: int
    dx: float
    lg: int
    quad_idx: int
    transform: float

    def __init__(
        self, subgrid_meta, num_refine_levels, min_gridsize, use_2d_flow, refinements
    ):
        pixel_size = subgrid_meta["pixel_size"]
        min_num_pix = min_gridsize / pixel_size
        if min_num_pix % 2 == 0:
            self.lgrmin = int(min_num_pix)
        else:
            raise SchematisationError(
                f"Smallest 2D grid cell does not contain an even number of pixels. "
                f"Smallest 2D grid cell size: {min_gridsize}m. Pixel size: {pixel_size}m."
            )

        # Maximum number of active grid levels in quadtree.
        self.kmax = reduce_refinement_levels(refinements, num_refine_levels)
        self.lgrmin *= 2 ** (num_refine_levels - self.kmax)

        # Array with cell widths at every active grid level [0:kmax]
        self.mmax = np.empty((self.kmax,), dtype=np.int32, order="F")
        # Array with row dimensions of every active grid level [0:kmax].
        self.nmax = np.empty((self.kmax,), dtype=np.int32, order="F")
        # Array with cell widths at every active grid level [0:kmax].
        self.dx = np.empty((self.kmax,), dtype=np.float64, order="F")
        self.origin = (subgrid_meta["bbox"][0], subgrid_meta["bbox"][1])
        self.pixel_size = subgrid_meta["pixel_size"]

        lvl_multiplr = 2 ** np.arange(self.kmax - 1, -1, -1, dtype=np.int32)

        # Determine number of largest cells that fit over subgrid extent.
        max_pix_largest_cell = self.min_cell_pixels * lvl_multiplr[0]
        max_large_cells_col = int(
            math.ceil(subgrid_meta["width"] / (max_pix_largest_cell))
        )
        max_large_cells_row = int(
            math.ceil(subgrid_meta["height"] / (max_pix_largest_cell))
        )

        # Calculate grid level dimensions.
        self.mmax[:] = max_large_cells_col * lvl_multiplr
        self.nmax[:] = max_large_cells_row * lvl_multiplr
        self.dx[:] = self.lgrmin * lvl_multiplr[::-1] * subgrid_meta["pixel_size"]

        # Array with dimensions of smallest active grid level and contains
        # map of active grid level for each quadtree cell.
        self.lg = self._apply_refinements(refinements)

        # Array with dimensions of smallest active grid level and contains
        # idx of active grid level for each quadtree cell.
        self.quad_idx = np.empty(
            (self.mmax[0], self.nmax[0]), dtype=np.int32, order="F"
        )
        self.n_cells = np.array(0, dtype=np.int32, order="F")
        self.n_lines_u = np.array(0, dtype=np.int32, order="F")
        self.n_lines_v = np.array(0, dtype=np.int32, order="F")

        m_quadtree.make_quadtree(
            self.kmax,
            self.mmax,
            self.nmax,
            self.min_cell_pixels,
            use_2d_flow,
            subgrid_meta["area_mask"],
            self.lg,
            self.quad_idx,
            self.n_cells,
            self.n_lines_u,
            self.n_lines_v,
        )

        if self.n_cells.sum() == 0:
            raise SchematisationError(
                "There are no 2D cells because the raster contains no active pixels"
            )

    def __repr__(self):
        return f"<Quadtree object with {self.kmax} refinement levels and {self.n_cells} active computational cells>"  # NOQA

    @property
    def min_cell_pixels(self):
        """Returns minimum number of pixels in smallles computational_cell."""
        return self.lgrmin

    def _apply_refinements(self, refinements):
        """Set active grid levels for based on refinements and
        setting lg variable for refinement locations.

        Args:
          refinements (GridRefinements): all gridrefinements polygon and linestrings
            with refinement levels.

        Returns:
            lg (ndarray): Array with spatial refinement locations.
        """

        if refinements is None:
            return np.full(
                (self.mmax[0], self.nmax[0]), self.kmax, dtype=np.int32, order="F"
            )

        lg = refinements.rasterize(
            origin=self.origin,
            height=self.nmax[0],
            width=self.mmax[0],
            cell_size=self.dx[0],
            no_data_value=self.kmax,
        )
        return np.asfortranarray(lg.T)

    def get_nodes_lines(self, area_mask, node_id_counter, line_id_counter):
        """Compute 2D openwater Nodes based on computed Quadtree.

        Return:
          Nodes object
        """

        # Create all node arrays for filling in external Fortran routine.

        # We need a counter for all nodes.
        id_n = itertools.islice(node_id_counter, self.n_cells)
        nodk = np.empty((self.n_cells,), dtype=np.int32, order="F")
        nodm = np.empty((self.n_cells,), dtype=np.int32, order="F")
        nodn = np.empty((self.n_cells,), dtype=np.int32, order="F")
        bounds = np.empty((self.n_cells, 4), dtype=np.float64, order="F")
        coords = np.empty((self.n_cells, 2), dtype=np.float64, order="F")
        pixel_coords = np.empty((self.n_cells, 4), dtype=np.int32, order="F")

        # Node type is always openwater at first init
        node_type = np.full(
            (self.n_cells,), NodeType.NODE_2D_OPEN_WATER, dtype="O", order="F"
        )

        # Create all line array for filling in external Fortran routine
        total_lines = self.n_lines_u + self.n_lines_v

        # Line type is always openwater at first init
        kcu = np.full((total_lines,), LineType.LINE_2D_U, dtype="i4", order="F")
        kcu[self.n_lines_u : total_lines] = LineType.LINE_2D_V

        # Node connection array
        line = np.empty((total_lines, 2), dtype=np.int32, order="F")
        cross_pix_coords = np.full((total_lines, 4), -9999, dtype=np.int32, order="F")

        m_cells.set_2d_computational_nodes_lines(
            np.array([self.origin[0], self.origin[1]], dtype=np.float64),
            self.min_cell_pixels,
            self.kmax,
            self.mmax,
            self.nmax,
            self.dx,
            self.lg,
            nodk,
            nodm,
            nodn,
            self.quad_idx,
            bounds,
            coords,
            pixel_coords,
            area_mask,
            line,
            cross_pix_coords,
            self.n_lines_u,
            self.n_lines_v,
        )

        idx = line[np.arange(total_lines), np.argmin(nodk[line[:, :]], axis=1)]
        lik = nodk[idx]
        lim = nodm[idx]
        lin = nodn[idx]

        nodes = Nodes(
            id=id_n,
            node_type=node_type,
            nodk=nodk,
            nodm=nodm,
            nodn=nodn,
            bounds=bounds,
            coordinates=coords,
            pixel_coords=pixel_coords,
            has_dem_averaged=0,
        )

        if np.any(line[:, 0] == line[:, 1]):
            raise RuntimeError("Quadtree produced self-connected lines")

        lines = Lines(
            id=itertools.islice(line_id_counter, len(line)),
            kcu=kcu,
            line=line,
            lik=lik,
            lim=lim,
            lin=lin,
            cross_pix_coords=cross_pix_coords,
        )

        return nodes, lines



import pytest

from threedigrid_builder.interface import DictOut


@pytest.fixture(scope="session")
def dict_out(grid_all):
    with DictOut(path=None) as out:
        result = out.write(grid_all, geometry_format="wkt")

    return result


def test_nodes(dict_out):
    assert dict_out["nodes"]["id"].shape == (3,)
    assert dict_out["nodes"]["geometry"][0] == "POINT (1 1)"
    assert dict_out["nodes"]["node_type"][0] == "NODE_2D_OPEN_WATER"


def test_lines(dict_out):
    assert dict_out["lines"]["id"].shape == (5,)
    assert dict_out["lines"]["geometry"][0] == "LINESTRING (1 1, 2 2)"


def test_cells(dict_out):
    assert dict_out["cells"]["id"].shape == (1,)
    assert dict_out["cells"]["geometry"][0] == "POLYGON ((1 0, 1 1, 0 1, 0 0, 1 0))"


def test_breaches(dict_out):
    assert dict_out["breaches"]["id"].shape == (2,)
    assert dict_out["breaches"]["geometry"][0] == "POINT (0 2)"


def test_nodes_embedded(dict_out):
    assert dict_out["nodes_embedded"]["id"].shape == (2,)


def test_meta(dict_out):
    assert dict_out["meta"]["epsg_code"] == 28992


@pytest.mark.parametrize(
    "group,dataset,expected",
    [
        ("nodes", "id", 1),
        ("nodes", "dmax", 1.2),
        ("nodes_embedded", "embedded_in", 2),
        ("lines", "id", 1),
        ("lines", "dpumax", 1.2),
        ("lines", "node_1", 1),  # reference to node
        ("lines", "node_2", 2),  # reference to node
        ("lines", "cross_id1", 4),  # reference to cross section def id, not increased
        ("breaches", "id", 1),
        ("breaches", "line_id", 4),  # reference to line
    ],
)
def test_not_off_by_one(dict_out, group, dataset, expected):
    # dict does not contain a dummy element, but ids are increased by 1 for consistency
    # with the gridadmin
    assert dict_out[group][dataset][0] == expected



import shapely

from threedigrid_builder.base import Array


class DemAverageArea:
    id: int
    the_geom: shapely.Geometry


class DemAverageAreas(Array[DemAverageArea]):
    pass


def apply_dem_average_areas(nodes, cell_tree, dem_average_areas):
    """Set calculation type to node being DEM_AVERAGED in tables. When the
    dem_average_areas polygons intersect with a 2D cell the node will be dem_averaged
    during table preprocessing.

    Args:
        nodes (Nodes)
        cell_tree (STRTree)
        dem_average_areas (DemAverageAreas)
    """
    idx = cell_tree.query(dem_average_areas.the_geom, "intersects")
    nodes.has_dem_averaged[idx[1, :]] = 1



import itertools
from enum import IntEnum
from typing import Tuple

import numpy as np
import pytest
from numpy.testing import assert_equal

from threedigrid_builder.base import Array, replace, search, TooManyExist


class Animal(IntEnum):
    ANT = 1
    BEE = 2
    CAT = 3
    DOG = 4


class Record:
    id: int
    number: float
    yesno: bool
    name: str
    animal: Animal
    xyz: Tuple[float, float, float]


class Records(Array[Record]):
    """Records docstring."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.length_for_init_test = len(self.id)

    def does_it_work(self):
        return "yes"


def test_arrays_from_lists():
    records = Records(
        id=[0, 1],
        number=[5, 7],
        yesno=[True, False],
        name=["a", "b"],
        animal=[Animal.ANT, -9999],
        xyz=[[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
    )

    assert_equal(records.id, np.array([0, 1], dtype=np.int32))
    assert_equal(records.number, np.array([5.0, 7.0], dtype=np.float64))
    assert_equal(records.yesno, np.array([True, False], dtype=bool))
    assert_equal(records.name, np.array(["a", "b"], dtype=object))
    assert_equal(records.animal, np.array([1, -9999], dtype=np.int32))
    assert_equal(
        records.xyz,
        np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]], dtype=np.float64, order="F"),
    )


def test_arrays_from_arrays():
    records = Records(
        id=np.array([0, 1], dtype=np.uint16),
        number=np.array([5, 7], dtype=int),
        yesno=np.array([1, 0], dtype=int),
    )

    assert_equal(records.id, np.array([0, 1], dtype=np.int32))
    assert_equal(records.number, np.array([5.0, 7.0], dtype=np.float64))
    assert_equal(records.yesno, np.array([True, False], dtype=bool))


def test_empty_values():
    records = Records(id=[0, 1])

    assert_equal(records.number, np.array([np.nan, np.nan], dtype=np.float64))
    assert_equal(records.yesno, np.array([False, False], dtype=bool))
    assert_equal(records.name, np.array([None, None], dtype=object))
    assert_equal(records.animal, np.array([-9999, -9999], dtype=np.int32))
    assert_equal(records.xyz, np.full((2, 3), np.nan, dtype=np.float64))


def test_broadcast_scalar_values():
    records = Records(id=[0, 1], number=5.2)

    assert_equal(records.number, np.array([5.2, 5.2], dtype=np.float64))


def test_id_from_generator():
    records = Records(id=(x for x in [5, 7]))
    assert_equal(records.id, [5, 7])


def test_id_from_islice():
    records = Records(id=itertools.islice([5, 7, 8], 2))
    assert_equal(records.id, [5, 7])


def test_id_required():
    with pytest.raises(TypeError):
        Records(number=[5, 7])


def test_id_dimensionality():
    with pytest.raises(ValueError):
        Records(id=7)


def test_id_sorted():
    with pytest.raises(ValueError):
        Records(id=[7, 5])


def test_unqual_lengths():
    with pytest.raises(ValueError):
        Records(id=[1], number=[5, 7])


@pytest.mark.parametrize(
    "xyz", [[[1, 1, 1], [2, 2, 2]], np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]]), None]
)
def test_fortran_order(xyz):
    records = Records(id=[0, 1], xyz=xyz)

    assert records.xyz.flags["F_CONTIGUOUS"]
    assert records.xyz.shape == (2, 3)


def test_enum_invalid():
    with pytest.raises(ValueError):
        Records(id=range(5), animal=[1, 2, 3, 4, 5])


def test_python_attrs():
    assert Records.__name__ == Records.__qualname__ == "Records"
    assert Records.__doc__ == "Records docstring."


def test_init():
    assert Records(id=[1, 2]).length_for_init_test == 2


def test_len():
    assert len(Records(id=[1, 2])) == 2


def test_repr():
    assert repr(Records(id=[2, 3])) == "<Records object, Record array of length 2>"


@pytest.mark.parametrize(
    "id,expected", [(1, 0), ([5], [2]), ([5, 3], [2, 1]), ([1, 1, 1], [0, 0, 0])]
)
def test_id_to_index(id, expected):
    records = Records(id=[1, 3, 5])

    assert_equal(records.id_to_index(id, check_exists=True), expected)


def test_id_to_index_check_exists():
    records = Records(id=[1, 3, 5])

    with pytest.raises(KeyError) as e:
        records.id_to_index([1, 2, 3], check_exists=True)
        assert_equal(e.indices, [1])
        assert_equal(e.values, [2])

    with pytest.raises(KeyError) as e:
        records.id_to_index([1, 1, 2, 2], check_exists=True)
        assert_equal(e.indices, [2, 3])
        assert_equal(e.values, [2, 2])

    records.id_to_index([1, 1, 2, 2])  # no check, no error


@pytest.mark.parametrize(
    "id,expected",
    [
        (1, 3),
        ([2], [5]),
        ([1, 0], [3, 1]),
        ([1, 1, 1], [3, 3, 3]),
        (-9999, -9999),
        ([-9999], [-9999]),
        ([1, -9999], [3, -9999]),
    ],
)
def test_index_to_id(id, expected):
    records = Records(id=[1, 3, 5])

    assert_equal(records.index_to_id(id), expected)


def test_methods():
    assert Records(id=[1]).does_it_work() == "yes"


def test_concatenate():
    a = Records(id=[1], xyz=[[0, 0, 0]])
    b = Records(id=[2], xyz=[[1, 1, 1]])
    actual = a + b
    assert len(actual) == 2
    assert_equal(actual.id, [1, 2])
    assert_equal(actual.xyz, [[0, 0, 0], [1, 1, 1]])


def test_concatenate_inplace():
    a = Records(id=[1], xyz=[[0, 0, 0]])
    b = Records(id=[2], xyz=[[1, 1, 1]])
    a += b
    assert len(a) == 2
    assert_equal(a.id, [1, 2])
    assert_equal(a.xyz, [[0, 0, 0], [1, 1, 1]])


def test_reorder():
    records = Records(id=[5, 7], number=[3, 2])
    records.reorder([1, 0])

    assert_equal(records.number, [2, 3])  # reordered
    assert_equal(records.id, [5, 7])  # kept the same


def test_reorder_by():
    records = Records(id=[5, 7], number=[3, 2])
    records.reorder_by("number")

    assert_equal(records.number, [2, 3])  # reordered
    assert_equal(records.id, [5, 7])  # kept the same


@pytest.mark.parametrize(
    "mask,expected",
    [
        ([True, True], [2, 3]),
        ([False, True], [3]),
        ([True, False], [2]),
        ([False, False], []),
        ([0, 1], [2, 3]),
        ([0], [2]),
        ([1], [3]),
        ([], []),
    ],
)
def test_indexing(mask, expected):
    records = Records(id=[5, 7], number=[2, 3])
    assert_equal(records[mask].number, records.number[mask])


def test_replace():
    actual = replace(
        np.array([0, 9, -150, 9999999]),
        {
            0: 1,
            9: 9999999,
            -150: -200,
            9999999: 9,
            1: 6,
        },
    )
    expected = np.array([1, 9999999, -200, 9])

    assert_equal(actual, expected)


def test_search():
    actual = search(
        [1, 4, 5, 9, 2],
        [4, 1, 2],
        assume_ordered=False,
        check_exists=True,
    )

    assert_equal(actual, [1, 0, 4])


def test_search_ordered():
    actual = search(
        [1, 2, 4, 5, 9],
        [4, 1, 2],
        assume_ordered=True,
        check_exists=True,
    )

    assert_equal(actual, [2, 0, 1])


def test_search_missing():
    actual = search(
        [1, 4, 5, 9, 2],
        [4, 1, 10],
        assume_ordered=False,
        check_exists=False,
    )

    assert_equal(actual, [1, 0, -9999])


def test_search_bool_mask():
    actual = search(
        [1, 4, 2, 9, 2],
        [4, 1, 2],
        mask=[True, True, False, False, True],
        assume_ordered=False,
        check_exists=True,
    )

    assert_equal(actual, [1, 0, 4])


def test_search_int_mask():
    actual = search(
        [1, 4, 2, 9, 2],
        [4, 1, 2],
        mask=[0, 1, 4],
        assume_ordered=False,
        check_exists=True,
    )

    assert_equal(actual, [1, 0, 4])


def test_search_int_mask_ordered():
    actual = search(
        [1, 4, 2, 9, 2],
        [4, 1, 2],
        mask=[0, 4, 1],  # this orders a
        assume_ordered=True,
        check_exists=True,
    )

    assert_equal(actual, [1, 0, 4])


def test_search_int_mask_missing():
    actual = search(
        [1, 4, 2, 9, 2],
        [4, 1, 9],
        mask=[0, 1, 4],
        assume_ordered=False,
        check_exists=False,
    )

    assert_equal(actual, [1, 0, -9999])


def test_search_check_exists():
    with pytest.raises(KeyError) as e:
        search(
            [1, 2, 4, 5, 9],
            [4, 6, 1, 2, 3],
            assume_ordered=False,
            check_exists=True,
        )
        assert_equal(e.indices, [1, 4])
        assert_equal(e.values, [6, 3])


def test_search_check_exists_mask():
    with pytest.raises(KeyError) as e:
        search(
            [1, 2, 4, 5, 9],
            [4, 1],
            mask=[0, 1, 3],
            assume_ordered=False,
            check_exists=True,
        )
        assert_equal(e.indices, [0])
        assert_equal(e.values, [4])


def test_search_empty():
    actual = search(
        [],
        [4, 1, 2],
        assume_ordered=False,
        check_exists=False,
    )

    assert_equal(actual, [-9999, -9999, -9999])


def test_search_empty_check_exists():
    with pytest.raises(KeyError):
        search(
            [],
            [4, 1, 2],
            assume_ordered=False,
            check_exists=True,
        )


def test_search_empty_mask():
    actual = search(
        [0],
        [4, 1, 2],
        mask=[False],
        assume_ordered=False,
        check_exists=False,
    )

    assert_equal(actual, [-9999, -9999, -9999])


def test_search_empty_mask_check_exists():
    with pytest.raises(KeyError):
        search(
            [0],
            [4, 1, 2],
            mask=[False],
            assume_ordered=False,
            check_exists=True,
        )


def test_split_in_two():
    records = Records(id=[0, 1, 2, 3, 4, 5])

    actual_1, actual_2 = records.split_in_two([0, 1, 1, 2, 3, 2])
    assert_equal(actual_1, [0, 1, 3, 4])
    assert_equal(actual_2, [2, 5])


def test_split_in_two_empty():
    records = Records(id=[])

    actual_1, actual_2 = records.split_in_two([])
    assert len(actual_1) == 0
    assert len(actual_2) == 0


def test_split_in_two_unique():
    records = Records(id=[0, 1, 2])

    actual_1, actual_2 = records.split_in_two([1, 2, 7])
    assert_equal(actual_1, [0, 1, 2])
    assert_equal(actual_2, [])


def test_split_in_two_err():
    records = Records(id=[0, 1, 2, 3])

    with pytest.raises(TooManyExist) as e:
        records.split_in_two([5, 2, 5, 5])

    assert_equal(e.value.values, [5])


class RecordsWithScalarArg(Array[Record]):
    """Records docstring."""

    scalars = ("foo",)


def test_scalar_arg_init():
    records = RecordsWithScalarArg(id=[], foo="bar")
    assert records.foo == "bar"


def test_scalar_arg_init_typeerror():
    with pytest.raises(TypeError):
        RecordsWithScalarArg(id=[])


def test_scalar_arg_getitem():
    records = RecordsWithScalarArg(id=[1, 2], foo="bar")
    actual = records[:1]
    assert len(actual) == 1
    assert actual.foo == "bar"


def test_scalar_arg_add():
    a = RecordsWithScalarArg(id=[1, 2], foo="bar")
    b = RecordsWithScalarArg(id=[3, 4], foo="baz")
    actual = a + b
    assert len(actual) == 4
    assert actual.foo == "bar"


@pytest.mark.parametrize(
    "name,index,expected",
    [
        ("number", [1, 2, -9999], [2.3, 3.4, np.nan]),
        ("name", [1, 2, -9999], ["b", "c", None]),
        ("animal", [1, 2, -9999], [2, 3, -9999]),
        ("number", 2, 3.4),
    ],
)
def test_take(name, index, expected):
    records = Records(
        id=range(3), name=["a", "b", "c"], animal=[1, 2, 3], number=[1.2, 2.3, 3.4]
    )

    assert_equal(records.take(name, index), expected)


def test_empty_column():
    records = Records(id=[1, 2, 3])
    assert "number" not in records.__dict__
    arr = records.number
    assert records.__dict__["number"] is arr
    assert_equal(arr, np.nan)



import numpy as np
import pytest
from osgeo import gdal

from threedigrid_builder.interface import GDALInterface


def test_does_not_exist(tmp_path):
    with pytest.raises(Exception):
        with GDALInterface(tmp_path / "i-do-not-exist.tif"):
            pass


@pytest.mark.skipif(
    gdal.VersionInfo().startswith("304"),
    reason="this test is known two fail with GDAL 3.4",
)
def test_crs(dem_path, crs_wkt_28992_legacy):
    with GDALInterface(dem_path) as dem:
        assert dem.crs == crs_wkt_28992_legacy


def test_pixel_size(dem_path):
    with GDALInterface(dem_path) as dem:
        assert dem.pixel_size == 0.5


def test_read(dem_path):
    with GDALInterface(dem_path) as dem:
        result = dem.read()
        assert result["area_mask"].shape == (9517, 9726)
        assert result["area_mask"].dtype == np.int16
        assert np.count_nonzero(result["area_mask"] == 1) == 47180799
        assert result["pixel_size"] == 0.5
        assert result["width"] == 9517
        assert result["height"] == 9726
        assert result["bbox"] == (106314.0, 514912.0, 111072.5, 519775.0)



import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_equal

from threedigrid_builder.base import LineHalfs, Lines, Nodes
from threedigrid_builder.constants import ContentType

CH = ContentType.TYPE_V2_CHANNEL
CN = ContentType.TYPE_V2_CONNECTION_NODES


@pytest.fixture
def nodes():
    return Nodes(
        id=[1, 2, 3], coordinates=[(1, 1), (2, 2), (3, 3)], content_type=[CN, CN, -9999]
    )


@pytest.fixture
def lines():
    return Lines(
        id=[0, 1, 2],
        line=[(1, 2), (2, 3), (1, 3)],
        line_geometries=[None, shapely.linestrings([(5, 5), (6, 6)]), None],
        ds1d=[np.nan, 16.0, 2.0],
        ds1d_half=[np.nan, 8.0, 1.1],
        content_type=[CN, CH, CH],
    )


def test_for_connection_nodes(nodes: Nodes, lines: Lines):
    line_halfs = LineHalfs.for_connection_nodes(nodes, lines)
    line_halfs.reorder_by("node_id")
    assert isinstance(line_halfs, LineHalfs)
    assert_equal(line_halfs.node_id, [1, 1, 2, 2])
    assert_equal(line_halfs.line_idx, [0, 2, 0, 1])
    assert_equal(line_halfs.is_start, [1, 1, 0, 1])


def test_for_connection_nodes_with_line_mask(nodes: Nodes, lines: Lines):
    line_halfs = LineHalfs.for_connection_nodes(
        nodes, lines, line_mask=lines.content_type == CH
    )
    line_halfs.reorder_by("node_id")
    assert isinstance(line_halfs, LineHalfs)
    assert_equal(line_halfs.node_id, [1, 2])
    assert_equal(line_halfs.line_idx, [2, 1])
    assert_equal(line_halfs.is_start, [1, 1])


def test_for_connection_nodes_with_node_mask(nodes: Nodes, lines: Lines):
    line_halfs = LineHalfs.for_connection_nodes(nodes, lines, node_mask=nodes.id != 2)
    assert isinstance(line_halfs, LineHalfs)
    assert_equal(line_halfs.node_id, [1, 1])
    assert_equal(line_halfs.line_idx, [0, 2])
    assert_equal(line_halfs.is_start, [1, 1])


def test_getattr(lines: Lines):
    line_halfs = LineHalfs(lines=lines, id=[0, 1, 2], line_idx=[1, 2, 1])

    assert_equal(line_halfs.content_type, lines.content_type[[1, 2, 1]])


@pytest.mark.parametrize(
    "ufunc,expected",
    [
        (np.fmax.reduceat, [16.0, 16.0]),
        (np.fmin.reduceat, [2.0, 16.0]),
        (np.add.reduceat, [18.0, np.nan]),
        (lambda x, y: y, [0.0, 2.0]),
    ],
)
def test_reduce(lines: Lines, ufunc, expected):
    line_halfs = LineHalfs(
        lines=lines, id=range(4), node_id=[1, 1, 5, 5], line_idx=[1, 2, 1, 0]
    )
    actual = line_halfs.reduce(ufunc, [16.0, 2.0, 16.0, np.nan])

    assert_almost_equal(actual, expected)


def test_reduce_2dim(lines: Lines):
    line_halfs = LineHalfs(
        lines=lines, id=range(4), node_id=[1, 1, 5, 5], line_idx=[1, 2, 1, 0]
    )
    actual = line_halfs.reduce(
        np.fmax.reduceat, np.array([[1.0, 4.0], [3.0, 3.0], [5.0, 6.0], [7.0, 8.0]])
    )

    assert_almost_equal(actual, np.array([[3.0, 4.0], [7.0, 8.0]]))


def test_reduce_empty_2dim(lines: Lines):
    line_halfs = LineHalfs(lines=lines, id=[])
    actual = line_halfs.reduce(None, np.empty((0, 2)))

    assert actual.shape == (0, 2)


def test_reduce_empty(lines: Lines):
    line_halfs = LineHalfs(lines=lines, id=[])
    actual = line_halfs.reduce(None, [])

    assert actual.shape == (0,)


@pytest.fixture
def line_halfs_for_agg(lines):
    return LineHalfs(
        lines=lines, id=range(4), node_id=[1, 1, 5, 5], line_idx=[1, 2, 0, 1]
    )


def test_reduce_id(line_halfs_for_agg: LineHalfs):
    actual = line_halfs_for_agg.get_reduce_id()

    assert_equal(actual, [1, 5])


def test_reduce_id_empty(lines: Lines):
    line_halfs = LineHalfs(lines=lines, id=[])
    actual = line_halfs.get_reduce_id()

    assert actual.shape == (0,)


def test_nanmin(line_halfs_for_agg: LineHalfs):
    actual = line_halfs_for_agg.nanmin([16.0, 2.0, np.nan, 16.0])

    assert_almost_equal(actual, [2.0, 16.0])


def test_first(line_halfs_for_agg: LineHalfs):
    actual = line_halfs_for_agg.first([16.0, 2.0, np.nan, 16.0])

    assert_almost_equal(actual, [16.0, np.nan])


def test_sum(line_halfs_for_agg: LineHalfs):
    actual = line_halfs_for_agg.sum([16.0, 2.0, np.nan, 16.0])

    assert_almost_equal(actual, [18.0, np.nan])


def test_invert_level(lines: Lines):
    lines.invert_level_start_point[:] = [1, 2, 3]
    lines.invert_level_end_point[:] = [4, 5, 6]

    line_halfs = LineHalfs(
        lines=lines, id=range(3), line_idx=[0, 2, 1], is_start=[False, True, False]
    )

    assert_almost_equal(line_halfs.invert_level, [4, 3, 5])


def test_ds1d(lines: Lines):
    line_halfs = LineHalfs(
        lines=lines, id=[0, 1, 2], line_idx=[1, 2, 2], is_start=[True, True, False]
    )

    assert_almost_equal(line_halfs.ds1d, [8.0, 1.1, 0.9])



import numpy as np
import shapely

from threedigrid_builder.base import Array
from threedigrid_builder.constants import CalculationType, ContentType
from threedigrid_builder.grid import linear
from threedigrid_builder.grid.cross_section_locations import compute_bottom_level

__all__ = ["Channels"]


class Channel:
    id: int
    code: str
    the_geom: shapely.Geometry
    dist_calc_points: float
    connection_node_start_id: int
    connection_node_end_id: int
    calculation_type: CalculationType
    display_name: str
    zoom_category: int
    exchange_thickness: float
    hydraulic_conductivity_out: float
    hydraulic_conductivity_in: float


class Channels(Array[Channel], linear.BaseLinear):
    content_type = ContentType.TYPE_V2_CHANNEL

    def is_closed(self, content_pk):
        """Whether objects are 'closed' or 'open water'.

        This is relevant for 1D-2D connections.
        """
        return np.full(len(content_pk), False, dtype=bool)

    def get_1d2d_exchange_levels(self, content_pk, s1d, locations, **kwargs):
        """Compute the exchange level (dpumax) for 1D-2D flowlines.

        For channels 1D-2D exchange levels are interpolated between crosssection
        location bank levels.

        Args:
            content_pk (array of int): object ids for which to compute levels
            s1d (array of int): distance on the objects where to compute levels
            locations (CrossSectionLocations): for the bank_levels

        Returns:
            exchange levels a.k.a. dpumax (array of float)
        """
        return compute_bottom_level(content_pk, s1d, locations, self, "bank_level")

    @property
    def has_groundwater_exchange(self):
        with np.errstate(invalid="ignore"):
            return (
                (self.exchange_thickness > 0)
                & np.isfinite(self.hydraulic_conductivity_out)
                & np.isfinite(self.hydraulic_conductivity_in)
            )



import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_equal
from shapely.testing import assert_geometries_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import LineType


@pytest.fixture
def nodes():
    return Nodes(id=[1, 2, 3], coordinates=[(1, 1), (2, 2), (3, 3)])


@pytest.fixture
def lines():
    return Lines(
        id=[0, 1, 2],
        line=[(1, 2), (2, 3), (1, 3)],
        line_geometries=[None, shapely.linestrings([(5, 5), (6, 6)]), None],
        ds1d=[np.nan, 16.0, 2.0],
    )


def test_set_line_coords(nodes, lines):
    lines.set_line_coords(nodes)

    assert_equal(lines.line_coords, [[1, 1, 2, 2], [2, 2, 3, 3], [1, 1, 3, 3]])


def test_set_line_coords_skips_already_set(nodes, lines):
    lines.line_coords[0] = (9, 9, 10, 10)
    lines.line_coords[1] = (1, 1, 1, np.nan)  # should be recomputed
    lines.set_line_coords(nodes)

    assert_equal(lines.line_coords[:2], [[9, 9, 10, 10], [2, 2, 3, 3]])


def test_fix_line_geometries(lines):
    lines.line_coords = np.array(
        [[1, 1, 2, 2], [np.nan, np.nan, np.nan, np.nan], [1, 1, 3, 3]]
    )

    lines.fix_line_geometries()

    expected = shapely.linestrings(
        [[(1, 1), (2, 2)], [(5, 5), (6, 6)], [(1, 1), (3, 3)]]
    )
    assert_geometries_equal(lines.line_geometries, expected)


def test_fix_ds1d(lines):
    lines.line_geometries = shapely.linestrings(
        [[(1, 1), (2, 2)], [(5, 5), (6, 6)], [(1, 1), (3, 3)]]
    )
    lines.fix_ds1d()

    assert_almost_equal(lines.ds1d, [np.sqrt(2), 16.0, 2.0])


def test_set_discharge_coefficients(lines):
    lines.discharge_coefficient_positive[:] = 2.0
    lines.discharge_coefficient_negative[0] = 3.0
    lines.set_discharge_coefficients()

    assert_equal(lines.discharge_coefficient_positive, [2.0, 2.0, 2.0])
    assert_equal(lines.discharge_coefficient_negative, [3.0, 1.0, 1.0])


def test_sort_by_nodes(lines):
    lines.line[:] = [[4, 5], [6, 7], [1, 0]]
    lines.sort_by_nodes([1, 4])

    assert_equal(lines.line, [[1, 0], [6, 7], [4, 5]])


U = LineType.LINE_2D_U
V = LineType.LINE_2D_V
U_OBS = LineType.LINE_2D_OBSTACLE_U
V_OBS = LineType.LINE_2D_OBSTACLE_V
L1D2D = LineType.LINE_1D2D_SINGLE_CONNECTED_CLOSED


@pytest.mark.parametrize(
    "kcu,crest_levels,expected_kcu,expected_flod",
    [
        ([U, V, V], [0.1, 0.3], [U_OBS, V_OBS, V], [0.1, 0.3, np.nan]),
        ([U, U, V], [], [U, U, V], [np.nan, np.nan, np.nan]),
    ],
)
def test_set_2d_crest_levels(
    lines: Lines, kcu, crest_levels, expected_kcu, expected_flod
):
    lines.kcu[:] = kcu
    lines.set_2d_crest_levels(
        np.array(crest_levels), where=np.arange(len(crest_levels))
    )

    assert_equal(lines.kcu, expected_kcu)
    assert_equal(lines.flod, expected_flod)
    assert_equal(lines.flou, expected_flod)


def test_line_get_velocity_points(lines: Lines):
    lines.ds1d_half[1] = np.sqrt(2) * 0.5

    actual = lines.get_velocity_points([1])
    assert len(actual) == 1
    assert shapely.to_wkt(actual[0]) == "POINT (5.5 5.5)"



from dataclasses import fields

import numpy as np
import shapely

from threedigrid_builder.base import (
    is_int_enum,
    is_tuple_type,
    Lines,
    OutputInterface,
    search,
    unpack_optional_type,
)
from threedigrid_builder.base.surfaces import SurfaceMaps, Surfaces
from threedigrid_builder.constants import ContentType, LineType, NodeType
from threedigrid_builder.grid import (
    Grid,
    GridMeta,
    Obstacles,
    PotentialBreaches,
    QuadtreeStats,
)
from threedigrid_builder.grid.cross_section_definitions import CrossSections

try:
    import h5py
except ImportError:
    h5py = None

__all__ = ["GridAdminOut"]


NODE_TYPES_1D = (NodeType.NODE_1D_NO_STORAGE, NodeType.NODE_1D_STORAGE)

LINE_TYPES_1D = (
    LineType.LINE_1D_EMBEDDED,
    LineType.LINE_1D_ISOLATED,
    LineType.LINE_1D_CONNECTED,
    LineType.LINE_1D_LONG_CRESTED,
    LineType.LINE_1D_SHORT_CRESTED,
    LineType.LINE_1D_DOUBLE_CONNECTED,
)

LINE_TYPES_1D2D = (
    LineType.LINE_1D2D_SINGLE_CONNECTED_CLOSED,
    LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
    LineType.LINE_1D2D_DOUBLE_CONNECTED_CLOSED,
    LineType.LINE_1D2D_DOUBLE_CONNECTED_OPEN_WATER,
)

LINE_TYPES_1D2D_GW = (LineType.LINE_1D2D_GROUNDWATER,)

LINE_TYPES_2D = (
    LineType.LINE_2D_U,
    LineType.LINE_2D_V,
    LineType.LINE_2D,
    LineType.LINE_2D_OBSTACLE,
    LineType.LINE_2D_OBSTACLE_U,
    LineType.LINE_2D_OBSTACLE_V,
)

# Some default settings for h5py datasets
HDF5_SETTINGS = {
    "compression": "gzip",  # more compatible than lzf and better compression
    "compression_opts": 1,  # the fastest, a 9 gives only 1% better compression
    "shuffle": True,  # helps another 3% and has almost no overhead
}


def field_to_h5(group, name, dtype, val, mode="attrs"):
    """Write a dataclass field to H5py attributes or datasets.

    None values are converted automatically to -9999 (int an IntEnum), NaN (float)
    and "" (str). bool values cannot be None. Unknown datatypes are skipped silently.
    """
    # backwards compat: epsg code is saved as string
    if name == "epsg_code":
        dtype = str
        val = str(val)
    # convert Optional[<type>] to <type>
    dtype = unpack_optional_type(dtype)
    # handle Tuple
    if is_tuple_type(dtype):
        if dtype.__args__[-1] is Ellipsis:
            shape = (len(val),)
        else:
            shape = (len(dtype.__args__),)
        dtype = dtype.__args__[0]
    else:
        shape = ()
    if dtype is bool:
        pass
    elif dtype is int or is_int_enum(dtype):
        dtype = np.int32
        if val is None:
            val = np.full(shape, -9999, dtype=dtype)
    elif dtype is float:
        dtype = np.float64
        if val is None:
            val = np.full(shape, np.nan, dtype=dtype)
    elif dtype is str:
        # Future: Write variable-length UTF-8
        # dtype = h5py.special_dtype(vlen=str)
        # But for calccore, we now need fixed-length ASCII
        if shape != ():
            raise RuntimeError("Cannot write string lists to HDF5.")
        if val is None:
            val = b""
        dtype = f"S{max(len(val), 1)}"
        if not isinstance(val, bytes):
            val = val.encode("utf-8", errors="ignore")
    else:
        return
    if mode == "attrs":
        group.attrs.create(name, data=val, shape=shape, dtype=dtype)
    elif mode == "datasets":
        group.create_dataset(name, data=val, shape=shape, dtype=dtype)
    else:
        raise ValueError(f"Unknown h5 write mode '{mode}'")


def dataclass_to_h5(group, datacls, mode="attrs"):
    """Write a dataclass to H5py attributes or datasets.

    None values are converted automatically to -9999 (int an IntEnum), NaN (float)
    and "" (str). bool values cannot be None. Unknown datatypes are skipped silently.
    """
    for field in fields(datacls):
        field_to_h5(group, field.name, field.type, getattr(datacls, field.name), mode)


def increase(arr):
    """Increase arr by one where arr is not -9999"""
    return np.add(arr, 1 * (arr != -9999), dtype=arr.dtype)


def to_bytes_array(arr, length):
    """Convert a list or array of strings to a numpy bytes array, utf-8 encoded."""
    result = np.zeros(len(arr), dtype=f"S{length}")
    for i, x in enumerate(arr):
        if x is None:
            continue
        elif isinstance(x, str):
            result[i] = x.encode("utf-8", errors="ignore")
        elif isinstance(x, bytes):
            result[i] = x
        else:
            raise TypeError(f"Unexpected type for bytes conversion, got '{type(x)}'")
    return result


class GridAdminOut(OutputInterface):
    def __init__(self, path):
        if not self.available():
            raise ImportError("Cannot write to HDF5 if h5py is not available.")
        super().__init__(path)

    @staticmethod
    def available():
        return h5py is not None

    def __enter__(self):
        self._file = h5py.File(self.path, mode="w")
        return self

    def __exit__(self, *args, **kwargs):
        self._file.close()

    def write(self, grid: Grid):
        self.write_meta(grid.meta)
        self.write_grid_counts(grid.nodes, grid.lines)
        self.write_quadtree(grid.quadtree_stats)
        self.write_nodes(grid.nodes)
        self.write_nodes_embedded(grid.nodes_embedded)
        self.write_lines(grid.lines, grid.cross_sections)
        self.write_pumps(grid.pumps)
        self.write_cross_sections(grid.cross_sections)
        self.write_obstacles(grid.obstacles)
        self.write_breaches(grid.breaches)
        if grid.meta.has_0d:
            self.write_surfaces(grid.surfaces, grid.surface_maps)

    def write_meta(self, meta: GridMeta):
        """Write the metadata to the gridadmin file.

        Metadata is stored in attrs and settings are stored in dedicated groups.

        Args:
            meta (GridMeta)
        """
        dataclass_to_h5(self._file, meta, "attrs")

        grid_settings = self._file.create_group("grid_settings")
        dataclass_to_h5(grid_settings, meta.grid_settings, mode="datasets")

        tables_settings = self._file.create_group("tables_settings")
        dataclass_to_h5(tables_settings, meta.tables_settings, mode="datasets")

    def write_grid_counts(self, nodes, lines: Lines):
        """Write the "meta" group in the gridadmin file.

        Args:
            nodes (Nodes)
            lines (Lines)

        Raises:
            ValueError if it exists already.
        """
        group = self._file.create_group("meta")

        # the number of nodes in several categories
        for dataset_name, node_types in [
            ("n1dtot", NODE_TYPES_1D),
            ("n2dtot", (NodeType.NODE_2D_OPEN_WATER)),
            ("ngrtot", (NodeType.NODE_2D_GROUNDWATER)),
            ("n1dobc", (NodeType.NODE_1D_BOUNDARIES)),
            ("n2dobc", (NodeType.NODE_2D_BOUNDARIES)),
            ("ngr2bc", (NodeType.NODE_2D_GROUNDWATER_BOUNDARIES)),
        ]:
            count = np.count_nonzero(np.isin(nodes.node_type, node_types))
            group.create_dataset(dataset_name, data=count, dtype="i4")

        # the number of lines in several categories
        # note: remove the 1D boundary lines, these are counted via n1dobc
        masked_line_kcu = lines.kcu[lines.is_1d_boundary != 1]
        for dataset_name, kcu_values in [
            ("liutot", (LineType.LINE_2D_U, LineType.LINE_2D_OBSTACLE_U)),
            ("livtot", (LineType.LINE_2D_V, LineType.LINE_2D_OBSTACLE_V)),
            ("l1dtot", LINE_TYPES_1D),
            ("l2dtot", LINE_TYPES_2D),
            ("lgrtot", LineType.LINE_2D_VERTICAL),
            ("infl1d", LINE_TYPES_1D2D),
            ("ingrw1d", LINE_TYPES_1D2D_GW),
        ]:
            count = np.count_nonzero(np.isin(masked_line_kcu, kcu_values))
            group.create_dataset(dataset_name, data=count, dtype="i4")

        # Groundwater lines from open water lines
        for dataset_name, kcu_values in [
            ("lgutot", (LineType.LINE_2D_U, LineType.LINE_2D_OBSTACLE_U)),
            ("lgvtot", (LineType.LINE_2D_V, LineType.LINE_2D_OBSTACLE_V)),
        ]:
            if self._file.attrs["has_groundwater_flow"]:
                count = np.count_nonzero(np.isin(masked_line_kcu, kcu_values))
            else:
                count = 0
            group.create_dataset(dataset_name, data=count, dtype="i4")

        # the number of unique boundaries (only 2D)
        for dataset_name, node_type in [
            ("nob2ds", NodeType.NODE_2D_BOUNDARIES),
            ("nob2dg", NodeType.NODE_2D_GROUNDWATER_BOUNDARIES),
        ]:
            count = len(np.unique(nodes.boundary_id[nodes.node_type == node_type]))
            group.create_dataset(dataset_name, data=count, dtype="i4")

    def write_quadtree(self, quadtree_statistics):
        """Write the "grid_coordinate_attributes" group in the gridadmin file.

        Args:
            quadtree (QuadTree)

        Raises:
            ValueError if it exists already.
        """
        if quadtree_statistics is None:
            # Write a dummy for pure 1D models (Issue 217)
            quadtree_statistics = QuadtreeStats(
                lgrmin=0,
                kmax=1,
                mmax=[1],
                nmax=[1],
                dx=[-10000000.0],
                dxp=-1000000.0,
                x0p=0.0,
                y0p=0.0,
            )

        group = self._file.create_group("grid_coordinate_attributes")
        dataclass_to_h5(group, quadtree_statistics, mode="datasets")

    def write_nodes_embedded(self, nodes_embedded):
        """Write the "nodes_embedded" group in the gridadmin file

        See write_nodes.
        """
        return self.write_nodes(nodes_embedded, group_name="nodes_embedded")

    def write_nodes(self, nodes, group_name="nodes"):
        """Write the "nodes" group in the gridadmin file

        Raises a ValueError if it exists already.

        Notes:
            Some datasets were 64-bit integers, but now they are saved as 32-bit integers.
            The following datasets were added: code.
            For floats (double) we use NaN instead of -9999.0 to denote empty.

        Args:
            nodes (Nodes)
        """
        group = self._file.create_group(group_name)
        shape = (len(nodes),)

        # Some convenient masks:
        is_cn = nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        is_2d = np.isin(
            nodes.node_type, [NodeType.NODE_2D_OPEN_WATER, NodeType.NODE_2D_GROUNDWATER]
        )
        is_manhole = nodes.manhole_id != -9999

        # Datasets that match directly to a nodes attribute:
        self.write_dataset(group, "id", nodes.id + 1)
        self.write_dataset(group, "code", to_bytes_array(nodes.code, 32))
        self.write_dataset(group, "node_type", nodes.node_type)
        self.write_dataset(group, "calculation_type", nodes.calculation_type)
        self.write_dataset(group, "coordinates", nodes.coordinates.T)
        self.write_dataset(group, "cell_coords", nodes.bounds.T)
        self.write_dataset(group, "nodk", nodes.nodk)
        self.write_dataset(group, "nodm", nodes.nodm)
        self.write_dataset(group, "nodn", nodes.nodn)
        self.write_dataset(group, "storage_area", nodes.storage_area)
        self.write_dataset(group, "dmax", nodes.dmax)
        self.write_dataset(group, "s1d", nodes.s1d)
        self.write_dataset(group, "embedded_in", nodes.embedded_in + 1)
        self.write_dataset(group, "boundary_id", nodes.boundary_id)
        self.write_dataset(group, "boundary_type", nodes.boundary_type)
        self.write_dataset(group, "has_dem_averaged", nodes.has_dem_averaged)
        self.write_dataset(group, "initial_waterlevel", nodes.initial_waterlevel)

        # content pk is only set for connection nodes, otherwise 0
        self.write_dataset(group, "content_pk", np.where(is_cn, nodes.content_pk, 0))
        # is_manhole is 1 for manholes, otherwise -9999
        self.write_dataset(
            group,
            "is_manhole",
            np.where(is_manhole, 1, -9999).astype("i4"),
        )
        self.write_dataset(group, "dimp", nodes.dimp)
        self.write_dataset(
            group, "bottom_level", np.where(is_manhole, nodes.dmax, np.nan)
        )
        self.write_dataset(group, "z_coordinate", nodes.dmax)

        # 2D stuff that is derived from 'bounds'
        pixel_width = np.full(len(nodes), -9999, dtype="i4")
        if is_2d.any():
            pixel_width[is_2d] = (
                nodes.pixel_coords[is_2d, 2] - nodes.pixel_coords[is_2d, 0]
            )

        self.write_dataset(group, "x_coordinate", nodes.coordinates[:, 0])
        self.write_dataset(group, "y_coordinate", nodes.coordinates[:, 1])
        self.write_dataset(group, "pixel_coords", nodes.pixel_coords.T)
        self.write_dataset(group, "pixel_width", pixel_width)
        self.write_dataset(
            group, "display_name", to_bytes_array(nodes.display_name, 64)
        )
        self.write_dataset(group, "zoom_category", nodes.zoom_category)
        self.write_dataset(group, "manhole_indicator", nodes.manhole_indicator)
        self.write_dataset(group, "shape", to_bytes_array(nodes.shape, 4))
        self.write_dataset(group, "drain_level", nodes.drain_level)
        self.write_dataset(group, "surface_level", nodes.surface_level)
        self.write_dataset(group, "width", nodes.width)

        # filled in threedi-tables:
        self.write_dataset(group, "sumax", np.full(shape, np.nan, dtype=np.float64))

    def write_lines(self, lines: Lines, cross_sections):
        """Write the "lines" group in the gridadmin file

        Raises a ValueError if it exists already.

        Notes:
            Some datasets were 64-bit integers, but now they are saved as 32-bit integers.
            The following datasets were added: ds1d, dpumax, flod, flou, cross1,
            cross2, cross_weight
            For floats (double) we use NaN instead of -9999.0 to denote empty.

        Args:
            lines (Lines)
        """
        group = self._file.create_group("lines")
        shape = (len(lines),)
        is_channel = lines.content_type == ContentType.TYPE_V2_CHANNEL

        l2d = np.isin(lines.kcu, (LineType.LINE_2D_U, LineType.LINE_2D_V))
        l2d_obstacle = np.isin(
            lines.kcu, (LineType.LINE_2D_OBSTACLE_U, LineType.LINE_2D_OBSTACLE_V)
        )
        # Datasets that match directly to a lines attribute:
        self.write_dataset(group, "id", lines.id + 1)
        self.write_dataset(group, "code", to_bytes_array(lines.code, 32))
        self.write_dataset(
            group, "display_name", to_bytes_array(lines.display_name, 64)
        )

        lines.kcu[l2d] = LineType.LINE_2D
        lines.kcu[l2d_obstacle] = LineType.LINE_2D_OBSTACLE
        self.write_dataset(group, "kcu", lines.kcu)
        calculation_type = np.full(shape, -9999, dtype="i4")
        calculation_type[is_channel] = lines.kcu[is_channel] + 100
        self.write_dataset(group, "calculation_type", calculation_type)
        self.write_dataset(group, "line", lines.line.T + 1)
        self.write_dataset(group, "cross_pix_coords", lines.cross_pix_coords.T)
        self.write_dataset(group, "s1d", lines.s1d)
        self.write_dataset(group, "ds1d_half", lines.ds1d_half)
        self.write_dataset(group, "ds1d", lines.ds1d)
        self.write_dataset(group, "lik", lines.lik)
        self.write_dataset(group, "lim", lines.lim)
        self.write_dataset(group, "lin", lines.lin)

        content_type = np.full(len(lines), b"", dtype="S10")
        for ind in np.where(lines.content_type != -9999)[0]:
            content_type_name = ContentType(lines.content_type[ind]).name
            content_type[ind] = content_type_name.lstrip("TYPE_").lower()[:10]

        self.write_dataset(group, "content_type", content_type)
        self.write_dataset(group, "content_pk", lines.content_pk)
        self.write_dataset(group, "dpumax", lines.dpumax)
        self.write_dataset(
            group, "invert_level_start_point", lines.invert_level_start_point
        )
        self.write_dataset(
            group, "invert_level_end_point", lines.invert_level_end_point
        )
        self.write_dataset(group, "flod", lines.flod)
        self.write_dataset(group, "flou", lines.flou)

        cross1 = lines.cross_id1.copy()
        cross1[cross1 != -9999] = np.digitize(
            cross1[cross1 != -9999], cross_sections.content_pk, right=True
        )
        cross2 = lines.cross_id2.copy()
        cross2[cross2 != -9999] = np.digitize(
            cross2[cross2 != -9999], cross_sections.content_pk, right=True
        )
        self.write_dataset(group, "cross1", increase(cross1))
        self.write_dataset(group, "cross2", increase(cross2))
        self.write_dataset(group, "frict_type1", lines.frict_type1)
        self.write_dataset(group, "frict_type2", lines.frict_type2)
        self.write_dataset(group, "frict_value1", lines.frict_value1)
        self.write_dataset(group, "frict_value2", lines.frict_value2)
        self.write_dataset(group, "cross_weight", lines.cross_weight)
        self.write_dataset(
            group, "hydraulic_resistance_in", lines.hydraulic_resistance_in
        )
        self.write_dataset(
            group, "hydraulic_resistance_out", lines.hydraulic_resistance_out
        )
        self.write_dataset(group, "line_coords", lines.line_coords.T)
        self.write_dataset(
            group,
            "discharge_coefficient",
            lines.discharge_coefficient_positive,
        )
        self.write_dataset(
            group,
            "discharge_coefficient_negative",
            lines.discharge_coefficient_negative,
        )
        self.write_dataset(
            group,
            "discharge_coefficient_positive",
            lines.discharge_coefficient_positive,
        )

        # Transform an array of linestrings to list of coordinate arrays (x,x,y,y)
        self.write_line_geometry_dataset(
            group, "line_geometries", lines.line_geometries
        )
        self.write_dataset(group, "zoom_category", lines.zoom_category)
        self.write_dataset(
            group, "connection_node_end_pk", lines.connection_node_start_id
        )
        self.write_dataset(
            group, "connection_node_start_pk", lines.connection_node_end_id
        )
        self.write_dataset(group, "crest_level", lines.crest_level)
        self.write_dataset(group, "crest_type", lines.crest_type)

        # Cross section on pipes, culverts, weirs, and orifices
        pipe_culvert = np.logical_or.reduce(
            (
                lines.content_type == ContentType.TYPE_V2_PIPE,
                lines.content_type == ContentType.TYPE_V2_CULVERT,
                lines.content_type == ContentType.TYPE_V2_WEIR,
                lines.content_type == ContentType.TYPE_V2_ORIFICE,
            )
        )
        # Data placeholders
        cross_section_width = np.full(len(lines), np.nan, dtype=np.float64)
        cross_section_height = np.full(len(lines), np.nan, dtype=np.float64)
        cross_section_shape = np.full(len(lines), -9999, dtype="i4")

        # Cross section definition ids, for pipe and culvert cross_id1 == cross_id2
        cross_ids = lines.cross_id1[pipe_culvert]

        # Put cross section data into placeholders at pipe and culvert indexes
        # Get cross section width, height, shape corresponding to cross section definition id
        np.put(
            cross_section_width,
            lines.id[pipe_culvert],
            cross_sections.width_1d[
                search(
                    cross_sections.content_pk, cross_ids, mask=None, assume_ordered=True
                )
            ],
        )
        np.put(
            cross_section_height,
            lines.id[pipe_culvert],
            cross_sections.height_1d[
                search(
                    cross_sections.content_pk, cross_ids, mask=None, assume_ordered=True
                )
            ],
        )
        np.put(
            cross_section_shape,
            lines.id[pipe_culvert],
            cross_sections.shape[
                search(
                    cross_sections.content_pk, cross_ids, mask=None, assume_ordered=True
                )
            ],
        )
        self.write_dataset(group, "cross_section_width", cross_section_width)
        self.write_dataset(group, "cross_section_height", cross_section_height)
        self.write_dataset(group, "cross_section_shape", cross_section_shape)
        self.write_dataset(group, "dist_calc_points", lines.dist_calc_points)
        self.write_dataset(group, "material", lines.material)
        self.write_dataset(group, "sewerage", lines.sewerage)
        self.write_dataset(group, "sewerage_type", lines.sewerage_type)
        has_friction_data = np.isin(
            lines.content_pk,
            [
                ContentType.TYPE_V2_PIPE,
                ContentType.TYPE_V2_CULVERT,
                ContentType.TYPE_V2_WEIR,
                ContentType.TYPE_V2_ORIFICE,
            ],
        )
        self.write_dataset(
            group,
            "friction_type",
            np.where(has_friction_data, lines.frict_type1, -9999),
        )
        self.write_dataset(
            group,
            "friction_value",
            np.where(has_friction_data, lines.frict_value1, np.nan),
        )
        self.write_dataset(group, "windshieldings", lines.windshieldings.T)

    def write_pumps(self, pumps):
        group = self._file.create_group("pumps")

        # Datasets that match directly to a lines attribute:
        self.write_dataset(group, "bottom_level", pumps.bottom_level)
        self.write_dataset(group, "code", to_bytes_array(pumps.code, 32))
        self.write_dataset(group, "content_pk", pumps.content_pk)
        self.write_dataset(group, "id", pumps.id + 1)
        self.write_dataset(group, "node1_id", increase(pumps.line[:, 0]))
        self.write_dataset(group, "node2_id", increase(pumps.line[:, 1]))
        self.write_dataset(group, "node_coordinates", pumps.line_coords.T)
        self.write_dataset(group, "capacity", pumps.capacity)
        self.write_dataset(
            group, "connection_node_end_pk", pumps.connection_node_end_id
        )
        self.write_dataset(
            group, "connection_node_start_pk", pumps.connection_node_start_id
        )
        self.write_dataset(group, "coordinates", pumps.coordinates.T)
        self.write_dataset(group, "lower_stop_level", pumps.lower_stop_level)
        self.write_dataset(group, "start_level", pumps.start_level)
        self.write_dataset(group, "type", pumps.type_)
        self.write_dataset(group, "upper_stop_level", pumps.upper_stop_level)
        self.write_dataset(
            group, "display_name", to_bytes_array(pumps.display_name, 64)
        )
        self.write_dataset(group, "zoom_category", pumps.zoom_category)

    def write_surfaces(self, surfaces: Surfaces, surface_maps: SurfaceMaps):
        # For now use 0 instead of NaN as fill value for surfaces
        # The calcore does not support NaN (for surfaces) yet
        default_fill_value = 0

        group = self._file.create_group("surface")
        self.write_dataset(group, "id", surfaces.id, fill=default_fill_value)
        self.write_dataset(group, "code", to_bytes_array(surfaces.code, 100))
        self.write_dataset(
            group, "display_name", to_bytes_array(surfaces.display_name, 250)
        )
        self.write_dataset(group, "function", to_bytes_array(surfaces.function, 64))

        self.write_dataset(
            group,
            "area",
            surfaces.area,
            fill=default_fill_value,
            fill_nan=default_fill_value,
        )
        self.write_dataset(
            group, "centroid_x", surfaces.centroid_x, fill=default_fill_value
        )
        self.write_dataset(
            group, "centroid_y", surfaces.centroid_y, fill=default_fill_value
        )
        self.write_dataset(
            group,
            "dry_weather_flow",
            surfaces.dry_weather_flow,
            fill=default_fill_value,
            fill_nan=default_fill_value,
        )
        self.write_dataset(
            group,
            "nr_of_inhabitants",
            surfaces.nr_of_inhabitants,
            fill=default_fill_value,
            fill_nan=default_fill_value,
        )
        self.write_dataset(group, "infiltration_flag", surfaces.infiltration_flag)
        self.write_dataset(
            group,
            "outflow_delay",
            surfaces.outflow_delay,
            fill=default_fill_value,
            fill_nan=default_fill_value,
        )
        self.write_dataset(
            group,
            "storage_limit",
            surfaces.storage_limit,
            fill=default_fill_value,
            fill_nan=default_fill_value,
        )

        if surfaces.surface_class is not None and np.any(
            surfaces.surface_class != None  # NOQA
        ):  # noqa
            # Impervious surfaces
            self.write_dataset(
                group, "surface_class", to_bytes_array(surfaces.surface_class, 128)
            )
            self.write_dataset(
                group,
                "surface_inclination",
                to_bytes_array(surfaces.surface_inclination, 64),
            )
            self.write_dataset(
                group,
                "surface_sub_class",
                to_bytes_array(surfaces.surface_sub_class, 128),
            )

        # Surface params
        self.write_dataset(group, "fb", surfaces.fb, fill=default_fill_value)
        self.write_dataset(group, "fe", surfaces.fe, fill=default_fill_value)
        self.write_dataset(group, "ka", surfaces.ka, fill=default_fill_value)
        self.write_dataset(group, "kh", surfaces.kh, fill=default_fill_value)

        # Surface maps (surface to connectionodes)
        self.write_dataset(group, "fac", surface_maps.fac, fill=default_fill_value)
        self.write_dataset(group, "nxc", surface_maps.nxc, fill=default_fill_value)
        self.write_dataset(group, "nyc", surface_maps.nyc, fill=default_fill_value)
        self.write_dataset(group, "pk", surface_maps.pk, fill=default_fill_value)

        # Note: +1 for Fortran 1-based indexing
        self.write_dataset(
            group, "cci", increase(surface_maps.cci), fill=default_fill_value
        )
        self.write_dataset(
            group, "imp", increase(surface_maps.imp), fill=default_fill_value
        )

    def write_cross_sections(self, cross_sections: CrossSections):
        if len(cross_sections) == 0:
            return
        group = self._file.create_group("cross_sections")

        # Datasets that match directly to a lines attribute:
        self.write_dataset(group, "id", cross_sections.id + 1)
        self.write_dataset(group, "code", to_bytes_array(cross_sections.code, 32))
        self.write_dataset(group, "shape", cross_sections.shape)
        self.write_dataset(group, "content_pk", cross_sections.content_pk)
        self.write_dataset(group, "width_1d", cross_sections.width_1d)
        self.write_dataset(group, "offset", cross_sections.offset)
        self.write_dataset(group, "count", cross_sections.count)
        self.write_dataset(group, "tables", cross_sections.tables.T, insert_dummy=False)

    def write_obstacles(self, obstacles: Obstacles):
        """For backwards compat, the group is named 'levees'"""
        if obstacles is None:
            return
        group = self._file.create_group("levees")

        self.write_dataset(group, "id", obstacles.id, insert_dummy=False)
        self.write_dataset(
            group, "crest_level", obstacles.crest_level, insert_dummy=False
        )
        self.write_dataset(
            group,
            "max_breach_depth",
            np.full(len(obstacles), fill_value=np.nan),
            insert_dummy=False,
        )
        self.write_line_geometry_dataset(
            group, "coords", obstacles.the_geom, insert_dummy=False
        )

    def write_breaches(self, breaches: PotentialBreaches):
        if breaches is None:
            return
        group = self._file.create_group("breaches")

        self.write_dataset(group, "id", breaches.id + 1)
        self.write_dataset(group, "levl", breaches.line_id + 1)
        self.write_dataset(group, "content_pk", breaches.content_pk)
        self.write_dataset(group, "levbr", breaches.maximum_breach_depth)
        self.write_dataset(group, "levmat", breaches.levee_material)
        self.write_dataset(
            group, "coordinates", shapely.get_coordinates(breaches.the_geom).T
        )
        self.write_dataset(group, "code", to_bytes_array(breaches.code, 32))
        self.write_dataset(
            group, "display_name", to_bytes_array(breaches.display_name, 64)
        )

    def write_dataset(
        self, group, name, values, fill=None, insert_dummy=True, fill_nan=None
    ):
        """Create the correct size dataset for writing to gridadmin.h5 and
        filling the extra indices with correct fillvalues.

        Args:
            group (hdf5 group): group to write to in hdf5 file.
            name (str): Name of dataset
            values (array): Values of dataset to write.
            fill: fillvalue with same dtype as values.
            fill_nane: (optional) fillvalue for nan values
        """

        if fill_nan is not None:
            if np.issubdtype(values.dtype, np.floating):
                values = values.copy()
                values[np.isnan(values)] = fill_nan

        if fill is None:
            if np.issubdtype(values.dtype, np.floating):
                fill = np.nan
            elif np.issubdtype(values.dtype, bytes):
                fill = b""
            else:
                fill = -9999
        assert insert_dummy in (True, False)

        if values.ndim == 1:
            ds = group.create_dataset(
                name,
                (values.shape[0] + int(insert_dummy),),
                dtype=values.dtype,
                fillvalue=fill,
                **HDF5_SETTINGS,
            )
            ds[int(insert_dummy) :] = values
            if insert_dummy and name == "id":  # set the ID of the dummy element
                ds[0] = 0
        elif values.ndim == 2:
            ds = group.create_dataset(
                name,
                (values.shape[0], values.shape[1] + int(insert_dummy)),
                dtype=values.dtype,
                fillvalue=fill,
                **HDF5_SETTINGS,
            )
            ds[:, int(insert_dummy) :] = values
        else:
            ValueError("Too many dimensions for values.")

    def write_line_geometry_dataset(self, group, name, data, insert_dummy=True):
        # Transform an array of linestrings to list of coordinate arrays (x,x,y,y)
        line_geometries = [shapely.get_coordinates(x).T.ravel() for x in data]
        if insert_dummy:
            line_geometries.insert(0, np.array([np.nan, np.nan]))
        # The dataset has a special "variable length" dtype
        try:
            vlen_dtype = h5py.vlen_dtype(np.dtype(float))
        except AttributeError:  # Pre h5py 2.10
            vlen_dtype = h5py.special_dtype(vlen=np.dtype(float))

        # insert line geometry data preserving its original type
        geometry_data = np.empty(len(line_geometries), dtype=object)
        geometry_data[:] = line_geometries

        group.create_dataset(
            name,
            data=geometry_data,
            dtype=vlen_dtype,
            **HDF5_SETTINGS,
        )
